from functools import wraps, partial
import numpy as np
import jax

def image_standardization(img):
    img = img - img.mean()
    axis = tuple(range(img.ndim-1))
    img = img / img.std(axis=axis, keepdims=True)

    if img.shape[-1] == 1:
        img = np.repeat(img, 3, axis=-1)
    elif img.shape[-1] == 2:
        img = np.c_[img, np.zeros_like(img[..., :1])]

    assert img.shape[-1] == 3

    return img


def gf_cycle(gf):
    @wraps(gf)
    def _wrapped(*args, **kwargs):
        while True:
            yield from gf(*args, **kwargs)

    return _wrapped


def gf_batch_(gf, *, batch_size, drop_remainder=True):
    @wraps(gf)
    def _wrapped(*args, **kwargs):
        stopping = False
        it = gf(*args, **kwargs)
        while not stopping:
            data = []
            for _ in range(batch_size):
                try:
                    data.append(next(it))
                except StopIteration:
                    stopping = True
                    break

            if len(data) > 0:
                if not drop_remainder or len(data) == batch_size:
                    data = jax.tree_util.tree_map(lambda *x: np.stack(x), *data)
                    yield data

    return _wrapped


def gf_batch(batch_size):
    return partial(gf_batch_, batch_size=batch_size)


from .convnext import ConvNeXt
from .integrators import FPN
from .lacss import Lacss
from .lpn import LPN
from .segmentor import Segmentor
from .unet import UNet
from .common import *
# from .video_integrator import VideoIntegrator



import numpy as np

def linear_assignment(cost_matrix, thresh):
    """ perform linear assignment 
    Args:
        cost_matrix: M x N cost matrix
        thresh: max cost for valid assignment

    Returns:
        matches: list of [idx_a, idx_b] pair indicating matches
        unmatched_a: all unmatched idx_a
        unmatched_b: all unmatched idx_b
    """
    import lap

    if cost_matrix.size == 0:
        return np.empty((0, 2), dtype=int), tuple(range(cost_matrix.shape[0])), tuple(range(cost_matrix.shape[1]))

    _, x, y = lap.lapjv(cost_matrix, extend_cost=True, cost_limit=thresh)

    matches = []
    for ix, mx in enumerate(x):
        if mx >= 0:
            matches.append([ix, mx])
    matches = np.asarray(matches)

    unmatched_a = np.where(x < 0)[0]
    unmatched_b = np.where(y < 0)[0]

    return matches, unmatched_a, unmatched_b



import jax


@jax.custom_vjp
def gradient_reversal(x):
    return x


def _gr_fwd(x):
    return x, None


def _gr_bwd(_, g):
    return (jax.tree_util.tree_map(lambda v: -v, g),)


gradient_reversal.defvjp(_gr_fwd, _gr_bwd)
""" A gradient reveral op. 

    This is a no-op during inference. During training, it does nothing
    in the forward pass, but reverse the sign of gradient in backward
    pass. Typically placed right before a discriminator in adversal 
    training.
"""



from __future__ import annotations

from functools import partial
from typing import Sequence, Tuple, Callable

import flax.linen as nn
import jax
import jax.numpy as jnp

from .common import DefaultUnpicklerMixin
from ..ops import non_max_suppression, distance_similarity
from ..typing import Array, ArrayLike

def generate_predictions(module, outputs):
    """
    Produce a list of proposal locations based on predication map, remove redundency with non_max_suppression
    """
    locations = outputs['pred_locs']
    scores = jax.nn.sigmoid(outputs['logits'])

    distance_threshold = module.nms_threshold
    output_size = module.max_output
    topk = module.pre_nms_topk
    score_threshold = module.min_score

    # sort and nms
    if topk <= 0 or topk > scores.size:
        topk = scores.size

    scores, indices = jax.lax.top_k(scores, topk)
    locations = locations[indices]

    threshold = 1 / distance_threshold / distance_threshold
    sel = non_max_suppression(
        scores,
        locations,
        output_size,
        threshold,
        score_threshold,
        return_selection=True,
    )

    idx_of_selected = jnp.argwhere(
        sel, size=output_size, fill_value=-1
    ).squeeze(-1)

    scores = jnp.where(idx_of_selected >= 0, scores[idx_of_selected], -1.0)
    locations = jnp.where(idx_of_selected[:, None] >= 0, locations[idx_of_selected], -1.0) 

    return dict(
        scores=scores,
        locations=locations,
    )


class LPN(nn.Module, DefaultUnpicklerMixin):
    """Location detection head

    Attributes:
        nms_threshold: non-max-supression threshold, if performing nms on detected locations.
        pre_nms_topk: max number of detections to be processed regardless of nms, ignored if negative
        max_output: number of detection outputs 
    """

    # network hyperparams
    feature_scale: int = 4
    dtype: Any = None

    # detection hyperparams
    nms_threshold: float = 8.0
    pre_nms_topk: int = -1
    max_output: int = 512
    min_score: float = 0.2

    @nn.compact
    def _block(self, feature: ArrayLike) -> dict:
        x = feature
        height, width = x.shape[:-1]

        logits = nn.Conv(1, (1, 1), dtype=self.dtype)(x)
        regressions = nn.Conv(2, (1, 1), dtype=self.dtype)(x)

        ref_locs = jnp.moveaxis(jnp.mgrid[:height, :width] + 0.5, 0, -1) * self.feature_scale
        locs = ref_locs + regressions * self.feature_scale


        return dict(
            regressions=regressions.reshape(-1, 2),
            logits = logits.reshape(-1),
            ref_locs=ref_locs.reshape(-1, 2),
            pred_locs=locs.reshape(-1, 2),
        )


    def __call__(self, feature: ArrayLike, mask: ArrayLike|None = None) -> dict:
        network_outputs = self._block(feature)
        predictions = generate_predictions(self, network_outputs)

        if mask is not None:
            pred_locs = tuple(jnp.floor(predictions["locations"]).transpose().astype(int))
            masked = mask[pred_locs]
            predictions['scores'] = jnp.where(masked, predictions['scores'], 0)
            predictions['locations'] = jnp.where(masked[:, None], predictions['locations'], -1)

        return dict(
            detector=network_outputs,
            predictions=predictions,
        )



from __future__ import annotations

import jax
import jax.numpy as jnp
import numpy as np
import optax

from xtrain import unpack_x_y_sample_weight

from ..ops import sorbel_edges, sorbel_edges_3d, coords_of_patches
from ..ops.patches import _get_patch_data
from .common import binary_focal_factor_loss, get_image_shape, mean_over_boolean_mask

EPS = jnp.finfo("float32").eps

# def _compute_edge(batch, prediction):
#     preds = prediction["predictions"]
#     instance_mask = preds["segmentation_is_valid"]
#     instance_logit = preds['segmentations']
#     instance_output = jax.nn.sigmoid(instance_logit)
#     img_shape  = get_image_shape(batch)
#     (zc, yc, xc), cmask = coords_of_patches(preds, img_shape)
#     cmask = cmask & instance_mask[:, None, None, None]

#     if img_shape[0] == 1: #2d
#         patch_edges = jnp.square(sorbel_edges(instance_output.squeeze(1)))
#         patch_edges = patch_edges[:, None, :, :]
#     else:
#         patch_edges = jnp.square(sorbel_edges(instance_output.reshape(-1, * instance_output.shape[2:])))
#         patch_edges = patch_edges.reshape(2, *instance_output.shape)
        
#     patch_edges = patch_edges.sum(axis=0) / 8
#     patch_edges = jnp.sqrt(jnp.clip(patch_edges, 1e-8, 1.0))  # avoid GPU error
#     combined_edges = jnp.zeros(img_shape, dtype=instance_logit.dtype)
#     combined_edges = combined_edges.at[zc, yc, xc].add(jnp.where(cmask, patch_edges, 0))
#     combined_edges = jnp.tanh(combined_edges)

#     return combined_edges

# def _compute_edge(batch, prediction):
#     preds = prediction["predictions"]
#     instance_logit = preds['segmentations']
#     instance_output = jax.nn.sigmoid(instance_logit)
#     img_shape  = get_image_shape(batch)
#     (zc, yc, xc), cmask = coords_of_patches(preds, img_shape)

#     if img_shape[0] == 1: #2d
#         patch_edges = jnp.square(sorbel_edges(instance_output.squeeze(1)) / 4)
#         patch_edges = patch_edges.sum(axis=0)
#         patch_edges = patch_edges[:, None, :, :]
#     else:
#         patch_edges = jnp.square(sorbel_edges_3d(instance_output) / 6)
#         patch_edges = patch_edges.mean(axis=0)
        
#     patch_edges = jnp.sqrt(jnp.clip(patch_edges, 1e-8, 1.0))  # avoid GPU error
#     combined_edges = jnp.zeros(img_shape, dtype=instance_logit.dtype)
#     combined_edges = combined_edges.at[zc, yc, xc].add(jnp.where(cmask, patch_edges, 0))
#     combined_edges = jnp.tanh(combined_edges)

#     return combined_edges

def _compute_edge(batch, prediction):
    preds = prediction["predictions"]
    instance_logit = preds['segmentations']
    instance_output = jax.nn.sigmoid(instance_logit)
    img_shape  = get_image_shape(batch)
    (zc, yc, xc), cmask = coords_of_patches(preds, img_shape)

    patch_edges = jnp.abs(sorbel_edges_3d(instance_output) / 6)
    combined_edges = jnp.zeros((3,) + img_shape, dtype=instance_logit.dtype)
    combined_edges = combined_edges.at[:, zc, yc, xc].add(jnp.where(cmask, patch_edges, 0))
    combined_edges = jnp.tanh(combined_edges)

    return combined_edges

def cks_boundry_loss(batch, prediction):
    """Cell border prediction consistency loss"""
    inputs, _, _ = unpack_x_y_sample_weight(batch)

    # lacss pred
    instance_edge = _compute_edge(batch, prediction)
    # if 'mask' in inputs:
    #     instance_edge = jnp.where(inputs['mask'], instance_edge, 0)

    # aux pred
    instance_edge_pred = jax.nn.sigmoid(prediction["predictions"]["edge_pred"])
    # instance_edge_pred = prediction["predictions"]["edge_pred"]
    if instance_edge_pred.ndim == 3:
        instance_edge_pred = instance_edge_pred[None, ]
        instance_edge = instance_edge[1:]

    instance_edge_pred = jnp.moveaxis(instance_edge_pred, -1, 0)
    assert instance_edge_pred.shape == instance_edge.shape

    if 'mask' in inputs:
        return optax.l2_loss(instance_edge_pred, instance_edge).mean(where=inputs['mask'])
    else:
        return optax.l2_loss(instance_edge_pred, instance_edge).mean()

def cks_segmentation_loss(batch, prediction, *, offset_sigma=10.0, offset_scale=2.0):
    """Image segmentation consistenct loss for the collaboraor model"""

    preds = prediction["predictions"]
    img_shape  = get_image_shape(batch)
    instance_mask = preds["segmentation_is_valid"]
    logits = jax.lax.stop_gradient(preds["segmentations"])

    ps_z, ps_y, ps_x = logits.shape[-3:]
    offset = np.moveaxis(np.mgrid[:ps_z, :ps_y, :ps_x], 0, -1)
    offset = offset - [(ps_z-1)/2, (ps_y-1)/2, (ps_x-1)/2]
    offset = offset * [3, 1, 1] #FIXME shoudl use variable z_scale
    offset = (offset ** 2).sum(axis=-1)
    offset = np.exp(- offset / 2 / offset_sigma / offset_sigma) * offset_scale

    logits = logits + jnp.asarray(offset)

    (cz, cy, cx), cmask = coords_of_patches(preds, img_shape)
    cmask = cmask & instance_mask[:, None, None, None]
    label = jnp.zeros(img_shape) - 1.0e6
    label = label.at[cz, cy, cx].max(jnp.where(cmask, logits, -1e6))

    fg_pred = jax.nn.sigmoid(preds["fg_pred"])
    fg_label = jax.nn.sigmoid(label)

    inputs, _, _ = unpack_x_y_sample_weight(batch)
    if 'mask' in inputs:
        fg_label = fg_label * inputs['mask']

    if fg_pred.ndim == 2:
        fg_pred = fg_pred[None, :, :]
    assert fg_pred.shape == fg_label.shape

    loss = (1 - fg_label) * fg_pred + fg_label * (1 - fg_pred)

    inputs, _, _ = unpack_x_y_sample_weight(batch)

    return loss.mean()


def aux_size_loss(batch, prediction, *, weight=0.01):
    """Auxillary loss to prevent model collapse"""
    preds = prediction["predictions"]
    inputs, labels, _ = unpack_x_y_sample_weight(batch)

    img_shape  = get_image_shape(batch)
    instance_mask = preds["segmentation_is_valid"]
    instances = jax.nn.sigmoid(preds["segmentations"])
    instance_mask = preds["segmentation_is_valid"]
    (cz, cy, cx), cmask = coords_of_patches(preds, img_shape)

    if labels is not None and ("gt_labels" in labels or "gt_masks" in labels):
        return None

    areas = jnp.sum(instances, axis=(-1, -2, -3), where=cmask)
    areas = jnp.clip(areas, EPS, 1e6)
    loss = jax.lax.rsqrt(areas / instances.shape[1]) * instances.shape[-1]
    loss = mean_over_boolean_mask(loss, instance_mask)

    return loss * weight


def supervised_segmentation_loss(batch, prediction):
    preds = prediction["predictions"]
    _, labels, _ = unpack_x_y_sample_weight(batch)

    if "gt_labels" in labels:

        mask = (labels["gt_labels"] > 0).astype("float32")

    else:

        mask = labels["gt_image_mask"].astype("float32")

    return optax.sigmoid_binary_cross_entropy(preds["fg_pred"], mask).mean()


def collaborator_segm_loss(batch, prediction, *, sigma, pi):
    preds = prediction["predictions"]
    _, labels, _ = unpack_x_y_sample_weight(batch)

    if not "fg_pred" in preds:
        return 0.

    if labels is None:
        labels = {}

    if "gt_image_mask" in labels or "gt_labels" in labels:
        return supervised_segmentation_loss(batch, prediction)
    else:
        return cks_segmentation_loss(
            batch, prediction, offset_sigma=sigma, offset_scale=pi
        )


def collaborator_border_loss(batch, prediction):
    preds = prediction["predictions"]
    _, labels, _ = unpack_x_y_sample_weight(batch)

    if not "edge_pred" in preds:
        return 0

    if labels is not None and ("gt_labels" in labels or "gt_masks" in labels):
        return 0

    else:
        return cks_boundry_loss(batch, prediction)



from __future__ import annotations

from typing import Any
from functools import partial
from copy import copy

import jax
import flax.linen as nn
import optax

from xtrain.utils import unpack_x_y_sample_weight, Inputs
from xtrain.strategy import VMapped, TrainState, LossLog
from xtrain.base_trainer import TrainIterator

from ..modules.common import PositionEmbedding2D, FFN
from ..ops import gradient_reversal

jnp = jax.numpy

def center_crop(x):
    h, w, _ = x.shape[-3:]
    return x[..., h//2-12:h//2+12, w//2-12:w//2+12, :]

def merge_h_w(x):
    return x.reshape(*x.shape[:-3], x.shape[-2]*x.shape[-3], x.shape[-1])


class Discriminator(nn.Module):
    @nn.compact    
    def __call__(self, features):
        pos = merge_h_w(PositionEmbedding2D()(features))

        x = merge_h_w(features)

        for _ in range(2):
            y = nn.LayerNorm()(x)
            y = nn.MultiHeadDotProductAttention(4, deterministic=True)(y+pos)
            x = x + FFN(deterministic=True)(y)

        y = nn.LayerNorm()(x)
        token = self.param("token", nn.initializers.normal(stddev=0.02), y.shape[:-2] + (1, y.shape[-1]))
        y = nn.MultiHeadDotProductAttention(4, deterministic=True)(token, y, y)

        dsc = nn.Dense(1)(y)
        return dsc


class LacssAdv(nn.Module):
    lacss: nn.Module

    @nn.compact
    def __call__(self, image, gt_locations=None, **kwargs):
        det_feature, seg_feature = gradient_reversal(self.lacss.get_features(image, None, deterministic=False))

        det_dsc = Discriminator()(det_feature)

        det_adv_loss = optax.sigmoid_binary_cross_entropy(
            det_dsc,
            jnp.ones_like(det_dsc) if gt_locations is None else jnp.zeros_like(det_dsc)
        ).mean()

        seg_dsc = Discriminator()(det_feature)
        seg_adv_loss = optax.sigmoid_binary_cross_entropy(
            seg_dsc,
            jnp.ones_like(seg_dsc) if gt_locations is None else jnp.zeros_like(seg_dsc)
        ).mean()
        
        # if gt_locations is None:
        #     detected = self.lacss.detector(det_feature)
        #     locations = (detected['predictions']['locations'][:128] / 4).astype(int)
        #     weights  = jnp.clip(detected['predictions']['scores'][:128] - 0.5, 0, 0.5)
        # else:
        #     locations = (gt_locations[:128]/ 4).astype(int)
        #     weights = jnp.where(locations[:, -1] > 0, 1.0, 0.0)

        # seg_patches, _ = self.lacss.segmentor._get_patch(seg_feature, locations)
        # seg_dsc = Discriminator()(seg_patches).squeeze(axis=(-1, -2))

        # seg_adv_loss = optax.sigmoid_binary_cross_entropy(
        #     seg_dsc,
        #     jnp.ones_like(seg_dsc) if gt_locations is None else jnp.zeros_like(seg_dsc)
        # )
        # seg_adv_loss = (seg_adv_loss * weights).sum() / (weights.sum() + 1e-6)

        output = dict(
            det_adv_loss = det_adv_loss,
            seg_adv_loss = seg_adv_loss,
        )

        return output


def _init_uda_it(it, inputs):
    model = LacssAdv(it.ctx.model)
    rngs = it.rngs.copy()
    rngs['params'] = jax.random.PRNGKey(123)
    params = model.init(rngs, inputs["image"][0])['params']
    params['lacss'] = it.parameters
    state = TrainState.create(
        apply_fn=model.apply,
        params=params,
        tx=it.ctx.optimizer,
    )
    it_ = copy(it)
    it_.train_state = state
    it_.loss_logs = [LossLog("det_adv_loss"), LossLog("seg_adv_loss")]
    # it_.loss_logs = [LossLog("det_adv_loss")]
    it_.variables = {}
    it_.has_aux = False
    it_.frozen = jax.tree_util.tree_map(lambda _:False, it_.parameters)
    it_.freeze("lacss")
    return it_

class AdversalUDA(VMapped):
    var_key:str = "adv"
    beta: float = 0.5

    @classmethod
    def train_step(
        cls,
        train_obj: TrainIterator,
        batch: Any,
    ) -> tuple[Any, TrainIterator]:
        inputs, label, _ = unpack_x_y_sample_weight(batch)

        if not cls.var_key in train_obj.variables:
            train_obj.variables[cls.var_key] = _init_uda_it(train_obj, inputs)

        uda_it = train_obj.variables[cls.var_key]
        del train_obj.variables[cls.var_key]

        uda_it.parameters['lacss'] = train_obj.parameters
        uda_grad, (prediction, uda_it) = jax.grad(cls.loss_fn, has_aux=True)(
            uda_it.parameters, uda_it, batch
        )
        uda_it.train_state = uda_it.train_state.apply_gradients(grads=uda_grad)
        del uda_it.train_state.params['lacss']

        if label is not None:
            grads, (inner_pred, train_obj) = jax.grad(cls.loss_fn, has_aux=True)(
                train_obj.parameters, train_obj, batch,
            )
            grads = jax.tree_util.tree_map(lambda a,b: a * cls.beta + b, uda_grad['lacss'], grads)
            prediction.update(inner_pred)
        else:
            grads = jax.tree_util.tree_map(lambda x: x * cls.beta, uda_grad['lacss'])

        train_obj.train_state = train_obj.train_state.apply_gradients(grads=grads)

        train_obj.variables[cls.var_key] = uda_it

        return prediction, train_obj



from .auxiliary import *
from .instance import *
from .common import *

def lpn_loss(batch, prediction):
    return prediction["losses"]["lpn_detection_loss"] + prediction["losses"]["lpn_localization_loss"]



from biopb.image.rpc_object_detection_pb2_grpc import ObjectDetection as Lacss
from biopb.image.rpc_object_detection_pb2_grpc import ObjectDetectionServicer as LacssServicer
from biopb.image.rpc_object_detection_pb2_grpc import ObjectDetectionStub as LacssStub
from biopb.image.rpc_object_detection_pb2_grpc import add_ObjectDetectionServicer_to_server as add_LacssServicer_to_server

from biopb.image.detection_request_pb2 import DetectionRequest
from biopb.image.detection_response_pb2 import DetectionResponse, ScoredROI
from biopb.image.bindata_pb2 import BinData
from biopb.image.detection_settings_pb2 import DetectionSettings
from biopb.image.image_data_pb2 import ImageData, Pixels, ImageAnnotation
from biopb.image.roi_pb2 import ROI, Rectangle, Mask, Mesh, Polygon, Point



from __future__ import annotations

import pathlib
from typing import Union, Mapping, Any, Sequence

import optax
from jax import Array
from jax.typing import ArrayLike

Optimizer = optax.GradientTransformation

PathLike = Union[str, pathlib.Path]

DataDict = Mapping[str, ArrayLike]

Patches = DataDict



import logging
import traceback
import threading
from concurrent import futures

import grpc
import numpy as np
import typer

from cellpose import models

from . import proto
from .common import decode_image, TokenValidationInterceptor

app = typer.Typer(pretty_exceptions_enable=False)

_MAX_MSG_SIZE=1024*1024*128
_TARGET_CELL_SIZE=30


def process_input(request: proto.DetectionRequest):
    pixels = request.image_data.pixels
    settings = request.detection_settings

    image = decode_image(pixels)

    physical_size = pixels.physical_size_x or 1

    if settings.HasField("cell_diameter_hint"):
        diameter = settings.cell_diameter_hint / physical_size

    else:
        diameter = _TARGET_CELL_SIZE / (settings.scaling_hint or 1.0)
    
    if image.shape[0] == 1: # 2D
        image = image.squeeze(0)

    if image.shape[-1] > 1:
        channels = [1, 2]
    else:
        channels = [0, 0]

    kwargs = dict(
        diameter = diameter,
        channels = channels,
    )

    return image, kwargs


def process_result(preds, image):
    import cv2
    from skimage.measure import regionprops

    response = proto.DetectionResponse()

    try:
        masks, flows, styles, _ = preds
    except:
        masks, flows, styles = preds

    for rp in regionprops(masks):
        mask = rp.image.astype("uint8")
        c, _ = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
        c = np.array(c[0], dtype=float).squeeze(1)
        c = c + np.array([rp.bbox[1] , rp.bbox[0]])
        c = c - 0.5

        scored_roi = proto.ScoredROI(
            score = 1.0,
            roi = proto.ROI(
                polygon = proto.Polygon(points = [proto.Point(x=p[0], y=p[1]) for p in c]),
            )
        )

        response.detections.append(scored_roi)

    return response


class CellposeServicer(proto.LacssServicer):

    def __init__(self, model):
        self.model = model
        self._lock = threading.RLock()

    def RunDetection(self, request, context):
        with self._lock:

            logging.info(f"Received message of size {request.ByteSize()}")

            try:
                image, kwargs = process_input(request)

                if image.ndim == 4:
                    raise ValueError("cellpose_server does not support 3D input")

                logging.info(f"received image {image.shape}")

                preds = self.model.eval(image,  **kwargs,)

                response = process_result(preds, image)

                logging.info(f"Reply with message of size {response.ByteSize()}")

                return response
            
            except ValueError as e:
                
                logging.error(repr(e))

                logging.error(traceback.format_exc())

                context.abort(grpc.StatusCode.INVALID_ARGUMENT, repr(e))

            except Exception as e:

                logging.error(repr(e))

                logging.error(traceback.format_exc())

                context.abort(grpc.StatusCode.UNKNOWN, f"prediction failed with error: {repr(e)}")



    def RunDetectionStream(self, request_iterator, context):
        with self._lock:
            request = proto.DetectionRequest()

            for next_request in request_iterator:

                if next_request.image_data.HasField("pixels"):
                    request.image_data.pixels.CopyFrom(next_request.image_data.pixels)

                if next_request.image_data.HasField("image_annotation"):
                    request.image_data.image_annotation.CopyFrom(next_request.image_data.image_annotation)
                
                if next_request.HasField("detection_settings"):
                    request.detection_settings.CopyFrom(next_request.detection_settings)
                
                if request.image_data.HasField("pixels"):
                    yield self.RunDetection(request, context)


@app.command()
def main(
    modeltype: str = "cyto3",
    port: int = 50051,
    workers: int = 10,
    ip: str = "0.0.0.0",
    local: bool = False,
    token: bool|None = None,
    debug: bool = False,
    compression: bool = True,
    gpu: bool = True,
    # max_image_size: int = 1088,
):
    logging.basicConfig(level=logging.DEBUG if debug else logging.INFO)

    print ("server starting ...")

    model = models.Cellpose(model_type = modeltype, gpu=gpu)

    if token is None:
        token = not local
    if token:
        import secrets

        token_str = secrets.token_urlsafe(64)

        print()
        print("COPY THE TOKEN BELOW FOR ACCESS.")
        print("=======================================================================")
        print(f"{token_str}")
        print("=======================================================================")
        print()
    else:
        token_str = None

    server = grpc.server(
        futures.ThreadPoolExecutor(max_workers=workers),
        compression=grpc.Compression.Gzip if compression else grpc.Compression.NoCompression,
        interceptors=(TokenValidationInterceptor(token_str),),
        options=(("grpc.max_receive_message_length", _MAX_MSG_SIZE),),
    )

    proto.add_LacssServicer_to_server(
        CellposeServicer(model), 
        server,
    )

    if local:
        server.add_secure_port(f"127.0.0.1:{port}", grpc.local_server_credentials())
    else:
        server.add_insecure_port(f"{ip}:{port}")

    logging.info(f"lacss_server: listening on port {port}")

    print ("server starting ... ready")

    server.start()
    server.wait_for_termination()


if __name__ == "__main__":
    app()




""" Various functions deals with segmentation pathces

    All functions here takes unbatched input. Use vmap to convert to batched data
"""
from __future__ import annotations

from functools import partial
from typing import Optional, Sequence

import jax
import jax.numpy as jnp
import numpy as np

from ..typing import *
from .boxes import box_iou_similarity
from .image import sub_pixel_samples

Shape = Sequence[int]

def gather_patches(source:Array, locations:ArrayLike, patch_size:int|tuple[int], *, padding_value=0):
    """ gather patches from an array

    Args:
        source: [D1..dk, ...]
        locations: [N, k] top-left coords of the patches, negative (out-of-bound) is ok
        patch_size: size of the patch. 
    Keyward Args:
        padding_value: value for out-of-bound locations

    returns:
        N patches [N, d1..dk, ...]
    """
    dim = locations.shape[-1]

    if locations.ndim != 2:
        raise ValueError(f"locations must be a 2d array, got shape {locations.shape}")

    if source.ndim < dim:
        raise ValueError(f"location has {dim} values, but the source data is less than {dim}-dim.")

    if isinstance(patch_size, int):
        patch_size = (patch_size,) * dim

    if len(patch_size) != dim:
        raise ValueError(f"length of patch_size does not match location shape.")

    src_shape = source.shape
    source = source.reshape(src_shape[:dim]+(-1,)) # [D1..Dk, B]

    slices = [slice(d) for d in patch_size]
    grid = jnp.mgrid[slices] + jnp.expand_dims(locations, axis=range(2, 2+dim)) # [N, k, d1..dk]
    grid = grid.swapaxes(0, 1) # [k, N, d1..dk]

    patches = source[tuple(grid)] # [N, d1..dk, B]
    max_d = jnp.expand_dims(jnp.asarray(src_shape[:dim]), axis=range(1, dim+2))
    valid_loc = ((grid >= 0) & (grid<max_d)).all(axis=0) # [N, d1..dk]
    patches = jnp.where(
        valid_loc[..., None],
        patches,
        padding_value,       
    ) # clear out-of-bound locations

    patches = patches.reshape(patches.shape[:-1] + src_shape[dim:])

    return patches

def _get_patch_data(pred):
    if isinstance(pred, dict):
        patches = pred["segmentations"].squeeze()

        if "segmentation_y_coords" in pred:# backward compatibility
            yc = pred["segmentation_y_coords"]
            xc = pred["segmentation_x_coords"]
        else: 
            ps = patches.shape[-1]
            yc, xc = jnp.mgrid[:ps, :ps]
            yc = yc + pred["segmentation_y0_coord"][:, None, None]
            xc = xc + pred["segmentation_x0_coord"][:, None, None]
    else:
        patches, yc, xc = pred
        patches = patches.squeeze() #FIXME 

    # not true for stacks
    # assert patches.ndim == yc.ndim

    return patches, yc, xc


def bboxes_of_patches(pred: dict, threshold: float = 0, *, image_shape:tuple|None = None, is2d:bool|None=None) -> jnp.ndarray:
    """Compute the instance bboxes from model predictions

    Args:
        pred: A model prediction dictionary:
        threshold: for segmentation, default 0

    Keyward Args:
        image_shape: if not None, the boxes are clipped to be within the bound
        is2d: whether to force 2d output

    Returns:
        bboxes: [n, 4] or [n, 6]
    """
    patches = pred["segmentations"]
    z0 = pred["segmentation_z0_coord"]
    y0 = pred["segmentation_y0_coord"]
    x0 = pred["segmentation_x0_coord"]

    n, d0, d1, d2 = patches.shape
    z_mask = jnp.any(patches > threshold, axis=(2,3))
    y_mask = jnp.any(patches > threshold, axis=(1,3))
    x_mask = jnp.any(patches > threshold, axis=(1,2))
    
    min_z = z_mask.argmax(axis=1)
    max_z = d0 - z_mask[:, ::-1].argmax(axis=1)

    min_y = y_mask.argmax(axis=1)
    max_y = d1 - y_mask[:, ::-1].argmax(axis=1)

    min_x = x_mask.argmax(axis=1)
    max_x = d2 - x_mask[:, ::-1].argmax(axis=1)

    min_z += z0
    max_z += z0
    min_y += y0
    max_y += y0
    min_x += x0
    max_x += x0

    if is2d is None:
        is2d = d0 == 1

    if image_shape is not None:
        if is2d:
            d = 1
            h, w = image_shape
        else:
            d, h, w = image_shape
        min_z = np.clip(min_z, 0, d-1)
        min_y = np.clip(min_y, 0, h-1)
        min_x = np.clip(min_x, 0, w-1)
        max_z = np.clip(max_z, 1, d)
        max_y = np.clip(max_y, 1, h)
        max_x = np.clip(max_x, 1, w)

    if is2d:
        bboxes = jnp.stack([min_y, min_x, max_y, max_x], axis=-1)
    else:
        bboxes = jnp.stack([min_z, min_y, min_x, max_z, max_y, max_x], axis=-1)

    is_valid = z_mask.any(axis=1)
    is_valid &= (max_z > min_z) & (max_y > min_y) & (max_x > min_x)
    bboxes = jnp.where(is_valid[:, None], bboxes, -1)

    return bboxes


def patches_to_label(
    pred: DataDict,
    input_size: Shape,
    *,
    mask: Optional[ArrayLike] = None,
    score_threshold: float = 0.5,
    threshold: float = 0,
) -> Array:
    """convert patch output to the image label.

    Args:
        pred: A model prediction dictionary
        input_size: shape of the input image. (H, W) or (D, H, W)
    Keyward Args:
        mask: boolean indicators masking out unwanted instances. Default is None (all cells)
        score_threshold: otional, min_score to be included. Default is .5.
        threshold: segmentation threshold.

    Returns:
        label: of the dimension input_size
    """
    input_size = tuple(input_size)
    if len(input_size) == 2:
        label = jnp.zeros((1,) + input_size)
    else:
        label = jnp.zeros(input_size)

    if mask is None:
        mask = pred["segmentation_is_valid"]
    else:
        mask &= pred["segmentation_is_valid"]
    if score_threshold > 0:
        mask &= pred["scores"] >= score_threshold

    assert label.ndim == 3, f'invalid input_size {input_size}'
    assert mask.ndim == 1

    patches = pred["segmentations"]
    (zz, yy, xx), cmask = coords_of_patches(pred, label.shape)

    idx = jnp.cumsum(mask) * mask
    idx = jnp.where(mask, idx.max() + 1 - idx, 0)

    pr = (patches > threshold).astype(int) * idx[:, None, None, None]

    label = label.at[zz, yy, xx].max(jnp.where(cmask, pr, 0))

    label = jnp.where(label, label.max() + 1 - label, 0)
    
    if len(input_size) == 2:
        label = label.squeeze(0)

    return label

_sampling_op = jax.vmap(partial(sub_pixel_samples, edge_indexing=True))

def rescale_patches(
    pred: DataDict, scale: float, *, transform_logits: bool = True
) -> tuple[Array, Array, Array]:
    """Rescale/resize instance outputs in a sub-pixel accurate way.
    If the input image was rescaled, this function take care of rescaling the predictions
    to the orginal coodinates.

    Args:
        pred: A model prediction dictionary
        scale: The scaling value. The function does not take a noop shortcut even if scale is 1.

    Returns: A tuple of three arrays
        patches: a 3D array of the rescaled segmentation patches. The array shape should be different from
            the orignal patches in model predition.
        yy: The y-coordinates of the patches in mesh-grid format
        xx: The x-coordinates of the patches in mesh-grid format
    """
    patch, yc, xc = _get_patch_data(pred)
    if transform_logits:
        patch = jax.nn.sigmoid(patch)

    new_y0 = (yc[:, 0, 0] + 0.5) * scale  # edge indexing in new scale
    new_x0 = (xc[:, 0, 0] + 0.5) * scale  # edge indexing in new scale
    ps = round(patch.shape[-1] * scale)

    yy, xx = jnp.mgrid[:ps, :ps]
    yy = (
        jnp.floor(new_y0[:, None, None]).astype(int) + yy
    )  # center indexing in new scale
    xx = (
        jnp.floor(new_x0[:, None, None]).astype(int) + xx
    )  # center indexing in new scale

    # edge indexing in old scale
    rel_yy = (yy + 0.5) / scale
    rel_yy -= yc[:, :1, :1]
    rel_xx = (xx + 0.5) / scale
    rel_xx -= xc[:, :1, :1]

    new_patch = _sampling_op(
        patch,
        jnp.stack([rel_yy, rel_xx], axis=-1),
    )

    return new_patch, yy, xx


def crop_and_resize_patches(pred:dict, bboxes, *, target_shape:tuple[int,...] = (48,48), convert_logits:bool=False):
    """ crop and rescale all instances to a target_size

    Args:
        pred: model predictions
        bboxes: optionally supply bboxes for cropping
    Keyward Args:
        target_shape: output shape. usually a 3-tuple but can be a 2-tuple if input is a 2D image.
        convert_logits: whether to convert the logits to probability
    
    Returns:
        Array [N] + target_shape.
    """
    from lacss.ops.image import sub_pixel_crop_and_resize

    patches = pred["segmentations"]
    z0 = pred["segmentation_z0_coord"]
    y0 = pred["segmentation_y0_coord"]
    x0 = pred["segmentation_x0_coord"]

    if bboxes.shape[-1] == 4:
        assert patches.shape[1] == 1, f"bboxes are for 2d but the patches are 3d"
        bboxes = jnp.c_[
            jnp.zeros_like(bboxes[:, 0]), 
            bboxes[:, :2],
            jnp.ones_like(bboxes[:, 0]), 
            bboxes[:, 2:],
        ]

    bboxes = bboxes - jnp.c_[z0, y0, x0, z0, y0, x0] # relative to instance crops

    target_shape = tuple(target_shape)
    if len(target_shape) == 2:
        assert patches.shape[1] == 1, f"target shape is 2d but the patches are 3d"
        target_size_ = (1,) + target_shape
    else:
        target_size_ = target_shape
    
    if convert_logits:
        patches = jax.nn.sigmoid(patches)

    segs = jax.vmap(partial(sub_pixel_crop_and_resize, output_shape=target_size_))(
        patches,
        bboxes,
    )
    
    if len(target_shape) == 2:
        segs = segs.squeeze(1)
        
    return segs

# def merge_patches(
#         preds:dict, 
#         input_size:tuple[int,...], 
#         *, 
#         min_score:float=0.5, 
#         reduction:str="sum",
#         convert_logits:bool=False,
# )->Array:
#     instance_mask = preds["segmentation_is_valid"]
#     if min_score > 0:
#         instance_mask &= preds["scores"] >= min_score
#     if len(input_size) == 2:
#         input_size = (1,) + tuple(input_size)

#     instances = preds["segmentations"]
#     if convert_logits:
#         instances = jax.nn.sigmoid(instances) #[N, sz, sy, sx]
    
#     sz, sy, sx = instances.shape[-3:]
#     pz, py, px = sz//2+1, sy//2+1, sx//2+1
#     zc, yc, xc = jnp.mgrid[:sz, :sy, :sx]
#     zc = zc + preds["segmentation_z0_coord"][:, None, None, None] + pz
#     yc = yc + preds["segmentation_y0_coord"][:, None, None, None] + py
#     xc = xc + preds["segmentation_x0_coord"][:, None, None, None] + px

#     merged = jnp.zeros(input_size, dtype=instances.dtype)
#     merged = jnp.pad(merged, [[pz, pz], [py, py], [px, px]])

#     if reduction == "sum":
#         merged = merged.at[zc, yc, xc].add(instances)
#     elif reduction == "max":
#         merged = merged.at[zc, yc, xc].max(instances)
#     else:
#         raise ValueError(f"invalid reduction {reduction}. should be 'sum' | 'max'")

#     merged = merged[pz:pz+sz, py:py+sy, px:px+sx]

#     return merged


def coords_of_patches(preds:dict, image_shape:tuple[int, ...])->tuple[Array, Array]:
    """ get the zyx coordinates of segmentations

    Args:
        preds: the model prediction dictionary
        image_shape: the original input image shape. (H, W) for 2d and (D, H, W) for 3d

    returns: a tuple of (coordinates, boolean_masks)
        coordinates: integer tensor of shape: (3,) + preds['segmentations'].shape
        boolan_masks: boolean tensor indicating whether the coordinate is a real one of padding.
    """
    locs = jnp.stack([ 
        preds["segmentation_z0_coord"],
        preds["segmentation_y0_coord"],
        preds["segmentation_x0_coord"],
    ]) #[3, N]

    sz, sy, sx = preds['segmentations'].shape[1:4] 
    if len(image_shape) == 2:
        assert sz == 1, f"2d image_shape but 3d predictions"
        image_shape = (1,) + tuple(image_shape)

    coords = jnp.mgrid[:sz, :sy, :sx] # [3, sz, sy, sx]
    coords = coords[:, None, ...] + locs[..., None, None, None] #[3, N, sz, sy, sx]
    mask = (coords >= 0).all(axis=0) & (jnp.moveaxis(coords, 0, -1) < jnp.asarray(image_shape)).all(axis=-1)
    mask = mask & preds["segmentation_is_valid"][:, None, None, None]

    return coords, mask



from typing import Mapping
# from .predict import Predictor

model_urls: Mapping[str, str] = {
    "lacss3-small": "https://huggingface.co/jiyuuchc/lacss3-small/resolve/main/lacss3-small",
    "lacss3-base": "https://huggingface.co/jiyuuchc/lacss3-base/resolve/main/lacss3-base",
    "lacss3-base-e": "https://huggingface.co/jiyuuchc/lacss3-base-e/resolve/main/lacss3-base-e",
    "lacss3-small-cellpose": "https://huggingface.co/jiyuuchc/lacss3-small-cellpose/resolve/main/lacss3-small-c",
    "lacss3-small-livecell": "https://huggingface.co/jiyuuchc/lacss3-small-livecell/resolve/main/lacss3-small-l",
    "lacss3-small-nips": "https://huggingface.co/jiyuuchc/lacss3-small-nips/resolve/main/lacss3-small-n",
}

model_urls["default"] = model_urls["lacss3-base"]




from __future__ import annotations

import math
from typing import Sequence, Callable, Any

import flax.linen as nn
import jax
import jax.numpy as jnp
from flax.linen import normalization

from ..typing import ArrayLike, Array
from .common import DefaultUnpicklerMixin, ChannelAttention

class Segmentor(nn.Module, DefaultUnpicklerMixin):
    """LACSS segmentation head.

    Attributes:
        n_cls: num of classes
        feature_scale: the spatail scale of the feature level
        instance_crop_size: Crop size for segmentation.
        pos_emb_shape: Dim of the learned position encoder.
    """
    feature_scale: int = 4
    instance_crop_size: int = 96
    patch_dim: int = 32
    sig_dim: int = 512
    pos_emb_shape: Sequence[int] = (16, 16, 4)
    dtype: Any = None

    @property
    def patch_size(self):
        return self.instance_crop_size // self.feature_scale

    def _get_patch(self, feature, locations)->tuple[Array, Array]:
        feature = nn.Conv(self.patch_dim, (3,3), dtype=self.dtype)(feature)

        patch_shape = (self.patch_size, self.patch_size)
        locations = locations.astype(int) - jnp.asarray(patch_shape) // 2  # so we can use loc for indexing
        coords = jnp.mgrid[:patch_shape[0], :patch_shape[1]] + locations[..., None, None] # N, 2, PS, PS

        limit = jnp.expand_dims(jnp.asarray(feature.shape[:-1]), (1,2))

        valid_locs = (coords >= 0).all(axis=1) & (coords < limit).all(axis=1) # N, PS, PS
        patches = jnp.where(
            valid_locs[..., None],
            feature[coords[:, 0], coords[:, 1]],
            0,
        )
        return patches, locations


    def _get_signiture(self, patches):
        bw = self.patch_size // 4
        n, _, _, n_ch = patches.shape
        x = jax.image.resize(
            patches[:, bw:-bw, bw:-bw, :],
            (n, 8, 8, n_ch),
            "linear",
        ).reshape(n, -1)

        x = nn.LayerNorm(dtype=self.dtype)(nn.Dense(self.sig_dim, dtype=self.dtype)(x))

        return x


    def _add_pos_encoding(self, patches, sig_vec):
        x = sig_vec
        x = nn.gelu(nn.Dense(self.sig_dim, dtype=self.dtype)(x))
        x = nn.gelu(nn.Dense(self.sig_dim, dtype=self.dtype)(x))

        x = nn.Dense(math.prod(self.pos_emb_shape), dtype=self.dtype)(x)
        x = jax.image.resize(
            x.reshape(-1, *self.pos_emb_shape),
            (patches.shape[0], self.patch_size, self.patch_size, self.pos_emb_shape[-1]),
            "linear",
        )

        encoding = nn.Dense(patches.shape[-1], use_bias=False, dtype=self.dtype)(x) #[n, ps, ps, dim]

        patches = jax.nn.relu(patches + encoding)

        return patches


    @nn.compact
    def __call__(self, feature: ArrayLike, locations: ArrayLike) -> dict:
        """
        Args:
            features: [H, W, C] features from the backbone.
            locations: [N, 2]

        Returns:
            A nested dictionary of values representing segmentation outputs.
              * segmentor/logits: all segment logits
              * predictions/segmentations: logits of designated class [N, D, ps, ps]
              * predictions/segmentation_y0_coord: y coord of patch top-left corner [N]
              * predictions/segmentation_x0_coord: x coord of patch top-left corner [N]
              * predictions/segmentation_is_valid: mask for valid patches [N]
        """
        locations = jnp.asarray(locations) / self.feature_scale

        x = jnp.asarray(feature)

        patches, patch_locs = self._get_patch(x, locations) # N, PS, PS, ch

        patches = ChannelAttention(dtype=self.dtype)(patches)

        patch_sigs = self._get_signiture(patches)

        patches = self._add_pos_encoding(patches, patch_sigs)

        logits = nn.ConvTranspose(1, (3, 3), strides=(2, 2), dtype=self.dtype)(patches)

        output_shape = (
            locations.shape[0],
            self.instance_crop_size,
            self.instance_crop_size
        )
        logits = jax.image.resize(logits.squeeze(-1), output_shape, "linear")
        instance_mask = locations[:, -1] >= 0
        patch_locs = jnp.where(
            instance_mask[:, None],
            patch_locs * self.feature_scale,
            -1,
        )

        return dict(
            segmentor=dict(
                patch_signitures=patch_sigs,
            ),
            predictions=dict(
                segmentations=logits[:, None, :, :], #[N, 1, ps, ps]
                segmentation_y0_coord=patch_locs[:, 0],
                segmentation_x0_coord=patch_locs[:, 1],
                segmentation_z0_coord=jnp.zeros_like(patch_locs[:, 0]),
                segmentation_is_valid=instance_mask,
            ),
        )



from __future__ import annotations

import numpy as np

def compute_mask_its(pred_contour, gt_label):
    import cv2
    from skimage.measure import regionprops
    from lacss.ops import box_intersection, box_area

    pred_polygons = pred_contour["pred_contours"]
    pred_bboxes = pred_contour["pred_bboxes"]
    pred_areas = np.array([ cv2.contourArea(ct.astype(int)) for ct in pred_polygons])

    if gt_label.shape[-1] == 1:
        gt_label = gt_label.squeeze(-1)
    dim = gt_label.ndim
    
    rps = regionprops(gt_label)
    gt_bboxes = np.stack([rp.bbox for rp in rps])
    gt_areas = np.stack([rp.area for rp in rps])

    box_its = box_intersection(pred_bboxes, gt_bboxes)

    mask_its = np.zeros_like(box_its, dtype=int)

    h, w = gt_label.shape
    def _get_its(pred_id, gt_id):
        polygon = pred_polygons[pred_id].astype(int)

        if polygon.size > 0:

            img = np.zeros(gt_label.shape, dtype="uint8")
            cv2.fillPoly(img, [polygon], 1)

            gt_coords = rps[gt_id].coords

            return np.count_nonzero(
                img[(gt_coords[:,0], gt_coords[:,1])]
            )

        else:
            return 0        

    ids = np.where(box_its > 0)
    mask_its[ids] = [_get_its(pid, gid) for pid, gid in zip(*ids)]

    return mask_its, pred_areas, gt_areas



from __future__ import annotations

from typing import Any
from functools import partial

import jax
import flax.linen as nn
import optax

from xtrain.utils import unpack_x_y_sample_weight, unpack_prediction_and_state, Inputs
from xtrain.strategy import VMapped, TrainState, LossLog
from xtrain.base_trainer import TrainIterator

from ..utils import deep_update

jnp=jax.numpy

def lacss_mt_train_fn(module, image, teacher_prediction):
    n_train_instances = 16

    seg_locs = teacher_prediction['predictions']['locations'][:n_train_instances]

    cutout_locs = jax.random.uniform(module.make_rng("dropout"), shape=seg_locs.shape) * 50 - 25 + seg_locs
    cutout_locs = jnp.where(seg_locs >=0 , cutout_locs.astype("int32"), 9999)
    image, _ = jax.lax.scan(
        lambda img, loc: (jax.lax.dynamic_update_slice(
            img, jnp.zeros([20,20,3], img.dtype), (loc[0]-10, loc[1]-10, 0),
        ), None),
        image, cutout_locs,        
    )

    x_det, x_seg = module.get_features(image, None, deterministic=False)
    outputs = module.detector(x_det)

    pred_cls_prob = jax.nn.sigmoid(outputs['detector']['logits'])
    gt_cls_prob =  jax.nn.sigmoid(teacher_prediction['detector']['logits'])
    mt_det_loss = pred_cls_prob * (1 - gt_cls_prob)  + (1 - pred_cls_prob) * gt_cls_prob
    mt_det_loss = mt_det_loss * jnp.abs(gt_cls_prob - .5)
    mt_det_loss = mt_det_loss.sum() /  jnp.abs(gt_cls_prob - .5).sum()

    pred_regr = outputs['detector']['regressions']
    gt_regr = teacher_prediction['detector']['regressions']
    mt_loc_loss = optax.l2_loss(pred_regr/4, gt_regr/4) * gt_cls_prob[:, None]
    mt_loc_loss = mt_loc_loss.sum() / gt_cls_prob[:, None].sum()

    segmentor_out = module.segmentor(x_seg, seg_locs)
    outputs = deep_update(outputs, segmentor_out)

    pred_segmentations = jax.nn.sigmoid(segmentor_out['predictions']['segmentations'])
    gt_segmentations = jax.nn.sigmoid(teacher_prediction['predictions']['segmentations'][:n_train_instances])
    # mt_seg_loss = pred_segmentations * (1 - gt_segmentations) + (1 - pred_segmentations) * gt_segmentations
    mt_seg_loss = optax.l2_loss(pred_segmentations, gt_segmentations) * jnp.abs(gt_segmentations - .5)
    mt_seg_loss = mt_seg_loss.sum() / jnp.abs(gt_segmentations - .5).sum()

    outputs['losses'] = dict(
        mt_det_loss = mt_det_loss,
        mt_loc_loss = mt_loc_loss,
        mt_seg_loss = mt_seg_loss,
    )

    return outputs


def lacss_mt_loss(params, teacher_params, train_obj, inputs):
    inputs = Inputs.from_value(inputs)
    batch_size = jax.tree_util.tree_leaves(inputs)[0].shape[0]

    inputs = inputs.update(
        rngs={
            name: jax.random.split(jax.random.fold_in(rng, train_obj.step), batch_size) 
            for name, rng in train_obj.rngs.items()
        },
        teacher_prediction=jax.vmap(Inputs.apply(
            train_obj.ctx.model.apply,
            dict(params=teacher_params),
        ))(inputs),
    )

    model_out = jax.vmap(Inputs.apply(
        nn.apply(lacss_mt_train_fn, train_obj.ctx.model),
        dict(params=params),
    ))(inputs)

    total_loss = sum([
        model_out['losses']['mt_det_loss'].sum(),
        model_out['losses']['mt_seg_loss'].sum(),
        model_out['losses']['mt_loc_loss'].sum(),
    ])

    return total_loss, model_out


class MeanTeacher(VMapped):
    alpha: float = 0.05
    var_key: str = "teacher"

    @classmethod
    def train_step(
        cls,
        train_obj: TrainIterator,
        batch: Any,
    ) -> tuple[Any, Any]:
        inputs, label, _ = unpack_x_y_sample_weight(batch)

        if not cls.var_key in train_obj.variables:
            train_obj.variables[cls.var_key] = dict(
                params = train_obj.parameters,
                mt_loss_log = [
                    LossLog("losses/mt_det_loss"),
                    LossLog("losses/mt_seg_loss"),
                    LossLog("losses/mt_loc_loss"),                    
                ],
            )

        teacher_vars = train_obj.variables.pop(cls.var_key)

        if label is not None:
            prediction, train_obj =  VMapped.train_step(train_obj, batch)

        else:
            grads, prediction = jax.grad(lacss_mt_loss, has_aux=True)(
                train_obj.parameters,
                teacher_vars['params'],
                train_obj,
                inputs
            )
            train_obj.train_state = train_obj.train_state.apply_gradients(grads=grads)
            for mt_losslog in teacher_vars["mt_loss_log"]:
                mt_losslog.update(batch, prediction)

        # update teacher params
        teacher_vars['params'] = jax.tree_map(
            lambda a, b: a * (1-cls.alpha) + b * cls.alpha,
            teacher_vars['params'],
            train_obj.parameters,
        )

        train_obj.variables[cls.var_key] = teacher_vars

        return prediction, train_obj



from __future__ import annotations

from typing import Dict, List, Sequence, Tuple, Union

import flax.linen as nn
import jax
import jax.numpy as jnp

from ..typing import *
from .common import *


class Bottleneck(nn.Module):
    n_filters: int
    strides: int = 1
    se_ratio: int = 16
    drop_rate: float = 0.0

    @nn.compact
    def __call__(self, inputs: jnp.ndarray, *, training: bool = None) -> jnp.ndarray:
        n_filters = self.n_filters
        strides = self.strides

        shortcut = inputs
        if strides > 1:
            shortcut = nn.Conv(
                n_filters * 4, (1, 1), strides=(strides, strides), use_bias=False
            )(shortcut)
            shortcut = nn.GroupNorm(num_groups=None, group_size=1)(shortcut)

        x = inputs
        x = nn.Conv(n_filters, (1, 1), use_bias=False)(x)
        x = nn.LayerNorm()(x)
        x = jax.nn.relu(x)
        x = nn.Conv(n_filters, (3, 3), strides=(strides, strides), use_bias=False)(x)
        x = nn.LayerNorm()(x)
        x = jax.nn.relu(x)
        x = nn.Conv(n_filters * 4, (1, 1), use_bias=False)(x)
        x = nn.LayerNorm()(x)
        x = jax.nn.relu(x)

        if self.se_ratio > 0:
            x = ChannelAttention(squeeze_factor=self.se_ratio)(x)

        deterministic = training is None or not training
        x = DropPath(self.drop_rate)(x, deterministic=deterministic)

        x = x + shortcut
        x = jax.nn.relu(x)

        return x


_RESNET_SPECS = {
    "50": [(64, 3), (128, 4), (256, 6), (512, 3)],
    "101": [(64, 3), (128, 4), (256, 23), (512, 3)],
    "152": [(64, 3), (128, 8), (256, 36), (512, 3)],
    "200": [(64, 3), (128, 24), (256, 36), (512, 3)],
    "270": [(64, 4), (128, 29), (256, 53), (512, 4)],
    "350": [(64, 4), (128, 36), (256, 72), (512, 4)],
    "420": [(64, 4), (128, 44), (256, 87), (512, 4)],
}


class ResNet(nn.Module):
    model_spec: Union[str, Sequence[Tuple[int, int]]] = "50"
    se_ratio: int = 16
    min_feature_level: int = 1
    stochastic_drop_rate = 0.0
    out_channels: int = 256
    patch_size: int = 2

    def __post_init__(self):

        if self.patch_size != 2 and self.patch_size != 4:
            raise ValueError(
                "Invalud patch_size {self.patch_size}. It shoule be either 2 or 4."
            )

    @nn.compact
    def __call__(self, x: jnp.ndarray, *, training: bool = None) -> dict:

        if self.patch_size == 2:
            x = jax.nn.relu(nn.Conv(24, (3, 3))(x))
            x = jax.nn.relu(nn.Conv(64, (3, 3))(x))
            encoder_out = [x]
        else:
            encoder_out = [x]
            x = jax.nn.relu(nn.Conv(24, (3, 3), strides=(2, 2))(x))
            x = jax.nn.relu(nn.Conv(64, (3, 3))(x))
            encoder_out.append(x)

        model_spec = self.model_spec
        if isinstance(model_spec, str):
            spec = _RESNET_SPECS[model_spec]
        else:
            spec = model_spec

        se_ratio = self.se_ratio
        stochastic_drop_rate = self.stochastic_drop_rate
        for i, (n_filters, n_repeats) in enumerate(spec):
            drop_rate = stochastic_drop_rate * (i + 2) / (len(spec) + 1)
            x = Bottleneck(n_filters, 2, se_ratio=se_ratio, drop_rate=drop_rate)(
                x, training=training
            )
            for _ in range(1, n_repeats):
                x = Bottleneck(n_filters, se_ratio=se_ratio, drop_rate=drop_rate)(
                    x, training=training
                )
            encoder_out.append(x)

        # l_min = self.min_feature_level
        # out_channels = self.out_channels

        # decoder_out = FPN(out_channels)(encoder_out[l_min:])

        # keys = [str(k) for k in range(len(encoder_out))]
        # encoder_out = dict(zip(keys, encoder_out))
        # decoder_out = dict(zip(keys[l_min:], decoder_out))

        return encoder_out



from __future__ import annotations

from typing import Sequence

from random import random
import numpy as np
from ..typing import ArrayLike

def _uniform_rand(min_v, max_v):
    return random() * (max_v - min_v) + min_v

def _init_boolean_mask(inputs):
    if "centroids" in inputs:
        return np.ones([len(inputs["centroids"])], dtype=bool)
    elif "bboxes" in inputs:
        return np.ones([len(inputs["bboxes"])], dtype=bool)
    else:
        return None

#FIXME try a np implementation
def _crop_and_resize(masks, boxes, target_shape):
    import jax

    from lacss.ops import sub_pixel_crop_and_resize
    return np.asarray(jax.vmap(
        sub_pixel_crop_and_resize,
        in_axes=(0, 0, None),
    )(masks, boxes, target_shape))


def _box_area(boxes):
    boxes = np.asarray(boxes)
    dim = boxes.shape[-1] // 2
    return np.abs(np.prod(boxes[..., dim:] - boxes[..., :dim], axis=-1))


def resize(inputs: dict, *, target_size: tuple[int, int]|tuple[int, int, int], use_jax:bool=False, p: float = 1.0) -> dict:
    """Resize image and labels

    Args:
        inputs: dict data:

            * image
            * image_mask
            * centroids
            * bboxes

    Keyword Args:
        target_size: target size
        p: probability of applying transformation
    """
    def _resize_img(img, shape):
        if use_jax:
            import jax
            if img.ndim > len(shape):
                shape = list(shape) + [img.shape[-1]]
            return np.asarray(jax.image.resize(img, shape, "linear"))
        else:
            from skimage.transform import resize
            return resize(img, shape)    

    def _resize(inputs):
        scaling = np.array(target_size) / inputs["image"].shape[:-1]
        inputs["image"] = _resize_img(inputs["image"], target_size)

        if "image_mask" in inputs:
            inputs["image_mask"] = _resize_img(inputs["image_mask"], target_size)
            # if inputs["image_mask"].dtype is np.dtypes.BoolDType:
            #     image_mask = image_mask >= inputs["image_mask"].max() * 0.5
            # inputs["image_mask"] = image_mask.astype(inputs["image_mask"].dtype)

        if "centroids" in inputs:
            inputs["centroids"] = inputs["centroids"] * scaling

        if "bboxes" in inputs:
            inputs["bboxes"] = inputs["bboxes"] * np.r_[scaling, scaling]

        return inputs

    if p < 1.0 and random() >= p:
        return inputs
    if tuple(target_size) == inputs['image'].shape[:-1]:
        return inputs
    else:
        return _resize(inputs.copy())


def rescale(
    inputs: dict,
    *,
    rescale: float|Sequence[float],
    use_jax:bool=False,
    p: float = 1.0,    
):
    """Rescale image and labels 

    Args:
        inputs: dict data:

            * image
            * image_mask
            * centroids
            * bboxes

    Keyword Args:
        rescale: rescaling factor
        p: probability of applying transformation
    """
    if p < 1.0 and random() >= p:
        return inputs

    imgsz = np.array(inputs['image'].shape[:-1])
    target_size = np.round(imgsz * rescale).astype(int)

    return resize(inputs, target_size=tuple(target_size), use_jax=use_jax)


def random_resize(
    inputs: dict,
    *,
    scaling: Sequence[float],
    keep_aspect_ratio: bool = False,
    p: float = 1.0,
) -> dict:
    """Resize image by a random amount

    Args:
        inputs: dict data:

            * image
            * image_mask
            * centroids

    Keyword Args:
        scaling: range of scale, e.g, [0.8, 1.5].
        keep_aspect_ratio: Whether to scale x/y the same amount.
        p: probability of applying transformation

    """

    if p < 1.0 and random() >= p:
        return inputs

    else:
        min_scale, max_scale = scaling
        dim = inputs['image'].ndim - 1

        if keep_aspect_ratio:
            scaling = [_uniform_rand(min_scale, max_scale)] * dim
        else:
            scaling = [_uniform_rand(min_scale, max_scale) for _ in range(dim)]

        target_size = np.array(inputs['image'].shape[:-1]) * scaling
        target_size = np.round(target_size).astype(int).tolist()

        return resize(inputs, target_size=target_size)


def crop_to_roi(
    inputs: dict,
    *,
    roi: Sequence[int] | ArrayLike,
    area_ratio_threshold: float = 1.0,
    clip_bboxes: bool = False,
    p: float = 1.0,
) -> dict:
    """Crop image to bounding-box ROI

    Args:
        inputs: dict

            * image
            * image_mask
            * centroids
            * bboxes
            * cell_masks

    Keyword Args:
        roi: Rectangle roi in yxyx (2d) or zyxzyx (3d) format
        area_ratio_threshold: remove instances if the bbox's relative remaining area is below this threshold
        clip_bboxes: for partially cropped cells, whether to clip bbox and cell_masks to be within the cropped region
        p: probability of applying transformation

    """
    roi = np.array(roi).astype(int)

    def _crop_to_roi(inputs):
        dim = inputs['image'].ndim - 1
        assert dim == 2 or dim == 3, f"invalid image shape {inputs['image'].sahpe}"
        span = roi[dim:] - roi[:dim]
        if (roi < 0).any() or (roi > inputs['image'].shape[:-1] * 2).any() or (span < 0).any():
            raise ValueError(f"invalid roi specified: {list(roi)}")

        if dim == 2:  
            inputs["image"] = inputs["image"][roi[0]:roi[2], roi[1]:roi[3], :]
        else:
            inputs["image"] = inputs["image"][roi[0]:roi[3], roi[1]:roi[4], roi[2]:roi[5],:]

        if "image_mask" in inputs:
            if dim == 2:
                inputs["image_mask"] = inputs["image_mask"][roi[0]:roi[2], roi[1]:roi[3]]
            else:
                inputs["image_mask"] = inputs["image_mask"][roi[0]:roi[3], roi[1]:roi[4], roi[2]:roi[5]]

        boolean_mask = _init_boolean_mask(inputs)

        if boolean_mask is None:
            return inputs

        if "bboxes" in inputs:
            bboxes = inputs["bboxes"] - np.r_[roi[:dim], roi[:dim]]
            areas = _box_area(bboxes)

            bboxes_clipped = np.clip(bboxes, 0, np.r_[span, span])
            areas_clipped = _box_area(bboxes_clipped)

            boolean_mask &= (areas_clipped / areas) >= area_ratio_threshold

            orig_bboxes = inputs["bboxes"]
            if clip_bboxes:
                inputs["bboxes"] = bboxes_clipped[boolean_mask]
            else:
                inputs["bboxes"] = bboxes[boolean_mask]

            if "centroids" in inputs:
                # for clipped cells, use center of bbox as new location
                ctrds = np.where(
                    (areas_clipped < areas)[:, None],
                    bboxes_clipped.reshape(-1, 2, dim).mean(axis=1),
                    inputs["centroids"] - roi[:dim],
                )
                inputs["centroids"] = ctrds[boolean_mask]
                        
            if "cell_masks" in inputs:
                inputs["cell_masks"] = inputs["cell_masks"][boolean_mask]
                if clip_bboxes:
                    target_shape = inputs["cell_masks"].shape[1:]

                    orig_bboxes = orig_bboxes[boolean_mask]
                    orig_r0 = orig_bboxes[:, :dim] - roi[:dim]
                    orig_size = orig_bboxes[:, dim:] - orig_bboxes[:, :dim]

                    new_bbox = inputs["bboxes"] - np.c_[orig_r0, orig_r0]
                    new_bbox = new_bbox / np.c_[orig_size, orig_size]
                    new_bbox = new_bbox * np.r_[target_shape, target_shape]

                    inputs["cell_masks"] = _crop_and_resize(
                        inputs["cell_masks"],
                        new_bbox,
                        target_shape,
                    )

        elif "centroids" in inputs:
            ctrds = inputs["centroids"] - roi[:dim]
            boolean_mask &= (ctrds >= 0).all(axis=-1) & (ctrds < span).all(axis=-1)

            inputs["centroids"] = ctrds[boolean_mask]


        return inputs

    if p < 1.0 and random() >= p:
        return inputs

    else:
        return _crop_to_roi(inputs.copy())


def random_crop(
    inputs: dict,
    *,
    target_size: Sequence[int] | ArrayLike,
    area_ratio_threshold: float = 1.0,
    clip_bboxes: bool = False,
    p: float = 1.0,
) -> dict:
    """Random crop to a set target size

    Args:
        inputs: dict data:

            * image
            * image_mask
            * centroids
            * bboxes
            * masks

    Keyword Args:
        target_size: the target size
        area_ratio_threshold: remove instances if the bbox's relative remaining area is below this threshold
        clip_bboxes: for partially cropped cells, whether to clip bbox and cell_masks to be within the cropped region
        p: probability of applying transformation
    """
    dim = inputs['image'].ndim - 1

    if len(target_size) != dim:
        raise ValueError(f"invalid target_size {target_size}")

    target_size = np.array(target_size).astype(int)
    if (target_size <= 0).any():
        raise ValueError(f"invalid target size {list(target_size)}")
    if (target_size > inputs['image'].shape[:-1]).any():
        raise ValueError(f"crop size {target_size} bigger than input size {inputs['image'].shape}")

    if p < 1.0 and random() >= p:

        return inputs

    else:
        r0 = np.array([
            int(random() * (inputs['image'].shape[k] - target_size[k])) 
            for k in range(dim)
        ])
        roi = np.r_[r0, r0 + target_size] 

        return crop_to_roi(
            inputs, roi = tuple(roi),
            area_ratio_threshold=area_ratio_threshold,
            clip_bboxes=clip_bboxes,
        )


def pad(
    inputs: dict,
    *,
    paddings: int|Sequence[int]|ArrayLike,
    constant_values: float = 0,
    p: float = 1.0,
) -> dict:
    """Pad image and labels.

    Args:
        inputs: dict data:

            * image
            * image_mask
            * centroids
            * bboxes

    Keyword Args:
        paddings: either a tuple or a single value. If latter, use the same padding for all axes
        constant_values: the value to fill the padded area
        p: probability of applying transformation
    """
    def _format_padding(p):
        try:
            pl, pr = p
        except:
            pl = pr = int(p) 
        return [pl, pr]

    if p < 1.0 and random() >= p:
        return inputs

    inputs = inputs.copy()
    dim = inputs['image'].ndim - 1
    try:
        iter(paddings)
    except:
        paddings = [int(paddings)] * dim

    paddings = [_format_padding(p) for p in paddings]

    inputs["image"] = np.pad(
        inputs["image"], list(paddings) + [[0, 0]],
        constant_values=constant_values,
    )

    if "image_mask" in inputs:
        inputs["image_mask"] = np.pad(
            inputs["image_mask"], paddings,
            constant_values=constant_values,
        )

    padding_left = [p[0] for p in paddings]
    if "centroids" in inputs:
        inputs["centroids"] = inputs["centroids"] + padding_left

    if "bboxes" in inputs:
        inputs["bboxes"] = inputs["bboxes"] + np.r_[padding_left, padding_left]
                                    
    return inputs


def pad_to_size(
    inputs: dict,
    *,
    target_size: Sequence[int] | ArrayLike,
    constant_values: float = 0,
    padding_type: str = "both",
    p: float = 1.0,
) -> dict:
    """Pad image and labels to a target size.

    Args:
        inputs: dict data:

            * image
            * image_mask
            * centroids
            * bboxes

    Keyword Args:
        target_size: target image size
        constant_values: the value to fill the padded area
        padding_type: "left", "right" or "both"
        p: probability of applying transformation
    """
    if not padding_type in ("left", "right", "both"):
        raise ValueError(f"unknown padding_type {padding_type}")

    if p < 1.0 and random() >= p:
        return inputs

    image = inputs['image']
    paddings = np.array(target_size).astype(int) - np.array(image.shape[:-1])
    if not (paddings >= 0).all():
        raise ValueError(f"target_size {target_size} smaller than input image {image.shape}")
    
    if padding_type == "left":
        paddings = np.c_[paddings, np.zeros_like(paddings)]
    elif padding_type == "right":
        paddings = np.c_[np.zeros_like(paddings), paddings]
    elif padding_type == "both":
        paddings_half = paddings // 2
        paddings = np.c_[paddings_half, paddings - paddings_half]
    else:
        raise ValueError(f"unknown padding type {padding_type}")

    return pad(inputs, paddings=paddings, constant_values=constant_values)


def random_crop_or_pad(
    inputs: dict,
    *,
    target_size: Sequence[int] | ArrayLike,
    constant_values: float = 0,
    padding_type: str = "both",
    area_ratio_threshold: float = 1.0,
    clip_bboxes: bool = False,
    p: float = 1.0,
) -> dict:
    """Random crop or pad image to specified target_size.

    Args:
        inputs: dict data:

            * image
            * image_mask
            * centroids
            * bboxes
            * cell_masks

    Keyword Args:
        target_size: target size
        constant_values: the value to fill the padded area
        padding_type: "left", "right" or "both"
        area_ratio_threshold: remove instances if the bbox's relative remaining area is below this threshold
        clip_bboxes: for partially cropped cells, whether to clip bbox and cell_masks to be within the cropped region
        p: probability of applying transformation
    """
    target_size = np.array(target_size).astype(int)
    if (target_size < 0).any():
        raise ValueError(f"target size cannot be negtive, got {target_size}")

    if p < 1.0 and random() >= p:
        return inputs

    else:
        inputs = pad_to_size(
            inputs, 
            target_size=np.maximum(inputs['image'].shape[:-1], target_size), 
            constant_values=constant_values,
            padding_type=padding_type,
        )

        inputs = random_crop(
            inputs, 
            target_size=target_size, 
            area_ratio_threshold=area_ratio_threshold,
            clip_bboxes=clip_bboxes,
        )

        return inputs



def sample_patch(
    inputs:dict, *, 
    target_size: Sequence[int] | ArrayLike,
    scaling: Sequence[float] | ArrayLike, 
    keep_aspect_ratio: bool = False, 
    constant_values: float = 0,
    padding_type: str = "both",
    area_ratio_threshold: float = 1.0,
    clip_bboxes: bool = False,
    p: float = 1.0
):

    if p < 1.0 and random() >= p:
        return inputs

    min_scale, max_scale = scaling
    dim = inputs['image'].ndim - 1

    assert len(target_size) == dim, f"target_size should have {dim} elements"

    if keep_aspect_ratio:
        scaling = [_uniform_rand(min_scale, max_scale)] * dim
    else:
        scaling = [_uniform_rand(min_scale, max_scale) for _ in range(dim)]

    patch_size = np.round(np.array(target_size) / scaling).astype(int)

    data = random_crop_or_pad(
        inputs, 
        target_size=patch_size.tolist(), 
        constant_values=constant_values, 
        padding_type=padding_type,
        area_ratio_threshold=area_ratio_threshold,
        clip_bboxes=clip_bboxes,
    )

    data = resize(data, target_size=target_size)

    return data   


def flip_left_right(inputs: dict, *, reformat_bboxes:bool=False, p: float = 1.0) -> dict:
    """Flip image left-right

    Args:
        inputs: dict data:

            * image
            * image_mask
            * centroids
            * bboxes
            * cell_masks

    Keyword Args:
        reformat_bboxes: whether to reformat bboxes so that x0 < x1, y0 < y1 etc.
        p: probability of applying transformation

    """
    if p < 1.0 and random() >= p:
        return inputs

    inputs = inputs.copy()
    img_shape = inputs['image'].shape[:-1]
    dim = len(img_shape)
    W = img_shape[-1]

    inputs["image"] = inputs["image"][..., ::-1, :]

    if "image_mask" in inputs:
        inputs["image_mask"] = inputs["image_mask"][..., ::-1]

    if "centroids" in inputs:
        inputs["centroids"][:, -1] = W - inputs["centroids"][:, -1]

    if "bboxes" in inputs:
        if reformat_bboxes:
            inputs["bboxes"][:, [-dim-1, -1]] = W - inputs["bboxes"][:, [-1, -dim-1]]
        else:
            inputs["bboxes"][:, [-dim-1, -1]] = W - inputs["bboxes"][:, [-dim-1, -1]]

    if "cell_masks" in inputs and reformat_bboxes:
        inputs["cell_masks"] = inputs["cell_masks"][..., ::-1]

    return inputs

def flip_up_down(inputs: dict, *, reformat_bboxes:bool=False, p: float = 1.0) -> dict:
    """Flip image up-down

    Args:
        inputs: dict data:

            * image
            * image_mask
            * centroids
            * bboxes
            * cell_masks

    Keyword Args:
        reformat_bboxes: whether to reformat bboxes so that x0 < x1, y0 < y1 etc.
        p: probability of applying transformation

    """
    if p < 1.0 and random() >= p:
        return inputs

    inputs = inputs.copy()
    img_shape = inputs['image'].shape[:-1]
    dim = len(img_shape)
    H = img_shape[-2]

    inputs["image"] = inputs["image"][..., ::-1, :, :]

    if "image_mask" in inputs:
        inputs["image_mask"] = inputs["image_mask"][..., ::-1, :]

    if "centroids" in inputs:
        inputs["centroids"][:, -2] = H - inputs["centroids"][:, -2]

    if "bboxes" in inputs: 
        if reformat_bboxes:
            inputs["bboxes"][:, [-2, -dim-2]] = H - inputs["bboxes"][:, [-dim-2, -2]]
        else:
            inputs["bboxes"][:, [-2, -dim-2]] = H - inputs["bboxes"][:, [-2, -dim-2]]

    if "cell_masks" in inputs and reformat_bboxes:
        inputs["cell_masks"] = inputs["cell_masks"][..., ::-1, :]

    return inputs


def flip_top_bottom(inputs: dict, *, reformat_bboxes:bool=False, p: float = 1.0) -> dict:
    """Flip image top and bottom

    Args:
        inputs: dict data:

            * image
            * image_mask
            * centroids
            * bboxes
            * cell_masks

    Keyword Args:
        reformat_bboxes: whether to reformat bboxes so that x0 < x1, y0 < y1 etc.
        p: probability of applying transformation

    """
    if p < 1.0 and random() >= p:
        return inputs

    inputs = inputs.copy()
    img_shape = inputs['image'].shape[:-1]
    dim = len(img_shape)
    if dim != 3:
        raise ValueError(f"called flip_top_bottom on 2d input")

    D = img_shape[-3]

    inputs["image"] = inputs["image"][::-1, :, :, :]

    if "image_mask" in inputs:
        inputs["image_mask"] = inputs["image_mask"][::-1, :, :]

    if "centroids" in inputs:
        inputs["centroids"][:, -3] = D - inputs["centroids"][:, -3]

    if "bboxes" in inputs: 
        if reformat_bboxes:
            inputs["bboxes"][:, [-3, -dim-3]] = D - inputs["bboxes"][:, [-dim-3, -3]]
        else:
            inputs["bboxes"][:, [-3, -dim-3]] = D - inputs["bboxes"][:, [-3, -dim-3]]

    if "cell_masks" in inputs and reformat_bboxes:
        inputs["cell_masks"] = inputs["cell_masks"][::-1, :, :]

    return inputs


def random_flip(inputs: dict, *, reformat_bboxes:bool=False, p: float = 1.0) -> dict:
    """Flip image on a random axis

    Args:
        inputs: dict data:

            * image
            * image_mask
            * centroids
            * bboxes
            * cell_masks

    Keyword Args:
        reformat_bboxes: whether to reformat bboxes so that x0 < x1, y0 < y1 etc.
        p: probability of applying transformation

    """
    if p < 1.0 and random() >= p:
        return inputs

    img_shape = inputs['image'].shape[:-1]
    dim = len(img_shape)

    flip_axis = int(random() * dim)
    if flip_axis == 0:
        return flip_left_right(inputs, reformat_bboxes=reformat_bboxes)
    elif flip_axis == 1:
        return flip_up_down(inputs, reformat_bboxes=reformat_bboxes)
    else:
        return flip_top_bottom(inputs, reformat_bboxes=reformat_bboxes)


def swapaxes(inputs: dict, *, axes:str="xy", p:float=1):
    """tranpose image on a random axis

    Args:
        inputs: dict data:

            * image
            * image_mask
            * centroids
            * bboxes
            * cell_masks

    Keyword Args:
        axes: "xy", "yz" or "xz". must be "xy" for 2d 
        p: probability of applying transformation
    """
    if p < 1.0 and random() >= p:
        return inputs
    
    inputs = inputs.copy()
    img_shape = inputs['image'].shape[:-1]
    dim = len(img_shape)

    assert dim ==2 or dim == 3, f"unexpected image dim {img_shape}"

    if (dim == 2 and axes != "xy") or not axes in ("xy", "xz", "yz"):
        raise ValueError(f"invalid axes {axes}")

    lut = {"x":2, "y":1, "z":0} if dim == 3 else {"x":1, "y":0}
    a1, a2 = (lut[x] for x in axes)
    
    inputs["image"] = np.swapaxes(inputs["image"], a1, a2)

    if "image_mask" in inputs:
        inputs["image_mask"] = np.swapaxes(inputs["image_mask"], a1, a2)

    if "centroids" in inputs:
        inputs["centroids"] = inputs["centroids"].copy()
        inputs["centroids"][:, [a1, a2]] = inputs["centroids"][:, [a2, a1]]

    if "bboxes" in inputs: 
        inputs["bboxes"] = inputs['bboxes'].copy()
        inputs["bboxes"][:, [a1, a2, a1+dim, a2+dim]] = inputs["bboxes"][:, [a2, a1, a2+dim, a1+dim]]

    if "cell_masks" in inputs:
        inputs["cell_masks"] = np.swapaxes(inputs["cell_masks"], a1+1, a2+1)

    return inputs

def random_swapaxes(inputs: dict, *, p:float=1):
    """tranpose image on a random axis

    Args:
        inputs: dict data:

            * image
            * image_mask
            * centroids
            * bboxes
            * cell_masks

    Keyword Args:
        p: probability of applying transformation
    """
    if p < 1.0 and random() >= p:
        return inputs

    img_shape = inputs['image'].shape[:-1]
    dim = len(img_shape)

    if dim == 2:
        return swapaxes(inputs)

    choices = ("xy", "yz", "xz")
    axes = choices[int(random() * 3)]
    return swapaxes(inputs, axes=axes)


def random_cut_out(inputs:dict, *, size:int, n:int = 1, p:float=1):
    if p < 1.0 and random() >= p:
        return inputs

    image = inputs['image'].copy()
    img_shape = image.shape[:-1]
    size = np.array(size).astype(int)

    sel = np.zeros([inputs['bboxes'].shape[0]], dtype=int)
    sel[:n] = 1
    np.random.shuffle(sel)
    sel_boxes = inputs['bboxes'][sel > 0]
    dim = sel_boxes.shape[-1]//2
    mins = np.minimum(sel_boxes[:,:dim], sel_boxes[:,dim:])
    maxs = np.maximum(sel_boxes[:,:dim], sel_boxes[:,dim:])
    c0s = np.random.randint(mins-size, maxs+size)

    for c0 in c0s:
        s = tuple(slice(max(0, a), min(c, b)) for a, b, c in zip(c0, c0+size, img_shape))
        image.__setitem__(s, 0)

    inputs['image'] = image

    return inputs
    

def mosaic(inputs: Sequence[dict]|dict):
    import jax
    if not isinstance(inputs, dict):
        assert len(inputs) == 4, f"mosaic requires a sequence of length 4, got{len(inputs)}"
        inputs = jax.tree_util.tree_map(lambda *x: x, *inputs) # ensure same pytree

    # tile
    images = inputs['image']
    img_shape = images[0].shape[:-1]
    image_tile = np.concatenate([
        np.concatenate(images[0:2], axis=-2),
        np.concatenate(images[2:4], axis=-2),
    ], axis=-3)
    del inputs['image']

    if "image_mask" in inputs:
        images = inputs['image_mask']
        image_mask_tile = np.concatenate([
            np.concatenate(images[0:2], axis=-1),
            np.concatenate(images[2:4], axis=-1),
        ], axis=-2)
        del inputs['image_mask']

    H, W = img_shape[-2:]
    offsets = np.array([
        [0, 0],
        [0, W],
        [H, 0],
        [H, W],
    ])
    if len(img_shape) == 3:
        offsets = np.c_[np.zeros([4, 1], dtype=int), offsets]
    
    if "centroids" in inputs:
        inputs["centroids"] = [
            locs + ofs
            for locs, ofs in zip(inputs["centroids"], offsets)
        ]
        
    if "bboxes" in inputs:
        inputs["bboxes"] = [
            box + np.r_[ofs, ofs]
            for box, ofs in zip(inputs["bboxes"], offsets)
        ]

    outputs = {k: np.concatenate(v) for k, v in inputs.items()}
    outputs['image'] = image_tile
    if "image_mask" in inputs:
        outputs["image_mask"] = image_mask_tile

    return outputs



import os

import jax.numpy as jnp
import numpy as np

def _to_str(p):
    return "".join(p.astype(int).reshape(-1).astype(str).tolist())


def format_predictions(pred, mask=None, encode_patch=True, threshold=0.5):
    """Produce more readable data from model predictions
    Args:
        pred: model output without batch dim
        mask: optional mask selecting cells
        threshold: float patch threshold
    Returns: dict(locations, scores, centroids, bboxes, encodings)
    """
    from lacss.ops import bboxes_of_patches

    patches = pred["instance_output"] >= threshold
    yc = pred["instance_yc"]
    xc = pred["instance_xc"]

    bboxes = bboxes_of_patches(pred)

    n_pixels = np.count_nonzero(patches, axis=(1, 2))
    centroids = jnp.stack(
        [
            (patches * yc).sum(axis=(1, 2)) / n_pixels,
            (patches * xc).sum(axis=(1, 2)) / n_pixels,
        ],
        axis=-1,
    )  # centroid

    outputs = dict(
        locations=pred["pred_locations"],
        scores=pred["pred_scores"],
        centroids=centroids,
        bboxes=bboxes,
    )

    is_valid = pred["instance_mask"].squeeze(axis=(-1, -2))
    is_valid &= patches.any(axis=(1, 2))  # no empty patches
    if mask is not None:
        is_valid &= mask

    outputs = {k: np.asarray(v)[is_valid] for k, v in outputs.items()}

    if encode_patch:

        encodings = []
        patches = np.asarray(patches)[is_valid]
        y0 = np.asarray(yc)[is_valid, 0, 0]
        x0 = np.asarray(xc)[is_valid, 0, 0]
        boxes = np.asarray(outputs["bboxes"]) - np.stack([y0, x0, y0, x0], axis=-1)

        for box, patch in zip(boxes, patches):
            roi = patch[box[0] : box[2], box[1] : box[3]]
            encodings.append(_to_str(roi))

        outputs["encodings"] = encodings

    return outputs


def show_images(imgs, locs=None, **kwargs):
    import matplotlib.patches
    import matplotlib.pyplot as plt

    fig, axs = plt.subplots(1, len(imgs), figsize=(4 * len(imgs), 5))
    if len(imgs) == 1:
        axs = [axs]

    for k, img in enumerate(imgs):
        axs[k].imshow(img, **kwargs)
        axs[k].axis("off")
        if locs is not None and locs[k] is not None:
            loc = np.round(locs[k]).astype(int)
            for p in loc:
                c = matplotlib.patches.Circle(
                    (p[1], p[0]), fill=False, edgecolor="white"
                )
                axs[k].add_patch(c)
    plt.tight_layout()


def dataclass_from_dict(klass, dikt):
    import dataclasses
    try:
        fieldtypes = {f.name: f.type for f in dataclasses.fields(klass)}
        return klass(**{f: dataclass_from_dict(fieldtypes[f], dikt[f]) for f in dikt})
    except:
        return dikt


def load_from_pretrained(pretrained: str | os.PathLike):
    """Load a saved model.

    Args:
        pretrained: The url to the saved model.

    Returns: A tuple (module, parameters) representing the model.
    """
    import os
    import pickle
    from flax.core.frozen_dict import unfreeze

    if os.path.isdir(pretrained):
        # directory are orbax checkpoint
        import orbax.checkpoint as ocp

        with open(os.path.join(os.path.dirname(pretrained), "..", "model.pkl"), "rb") as f:
            module = pickle.load(f)

        cp_dir = os.path.abspath(pretrained)
        if os.path.exists(os.path.join(cp_dir, "default")):
            params = ocp.StandardCheckpointer().restore(os.path.join(cp_dir, "default"))
        else:
            params = ocp.StandardCheckpointer().restore(cp_dir)

        params = ( params['train_state']['params'], )

    else:
        # uri or files were treated as pickled byte steam
        from .modules import Lacss

        if os.path.isfile(pretrained):
            with open(pretrained, "rb") as f:
                thingy = pickle.load(f)

        else:
            from urllib.request import Request, urlopen

            headers = {"User-Agent": "Wget/1.13.4 (linux-gnu)"}
            req = Request(url=pretrained, headers=headers)

            bytes = urlopen(req).read()
            thingy = pickle.loads(bytes)

        cfg, *params = thingy

        assert len(params) >= 1, f"No parameter found for model {cfg}"

        if isinstance(cfg, Lacss):
            module = cfg
        else:
            try:
                module = Lacss.from_config(cfg)
            except:
                raise RuntimeError("Cannot interpet {cfg} as an Lacss model")

    params = [ p['params'] if 'params' in p else p for p in params]

    if len(params) == 1:
        params = params[0]

    # for backward compatibility
    # if not "cnn" in params['backbone']:
    #     params['backbone']['cnn'] = params['backbone']['ConvNeXt_0']
    #     del params['backbone']['ConvNeXt_0']
    
    # if not "Scan_PatchOp_0" in params["segmentor"]:
    #     params = unfreeze(params)
    #     params['segmentor']['Scan_PatchOp_0']={}
    #     for k in ['ConvTranspose_0', 'Dense_0', 'Dense_1', 'Dense_2', 'Dense_3', 'Dense_4', 'Dense_5']:
    #         params['segmentor']['Scan_PatchOp_0'][k] = params['segmentor'][k]
    #         del params['segmentor'][k]

    return module, params


def make_label_continuous(label, dtype=None):
    """Relabel a label image so that the label values are continuous. It is assumed that the
    label starts with 0. If there is negative numbers in the input, they will be replaced by 0.

    Args:
        label: input label image

    Keyword Args:
        dtype: the dtype of the output image. default is to keep the dtype of the input.

    Returns:
        Relabeled image.
    """
    if not isinstance(label, np.ndarray) and dtype is None:
        raise ValueError(
            "A dtype must be specified if the input data is not a np array"
        )
    elif dtype is None:
        dtype = label.dtype

    label = np.asarray(label, dtype=dtype)

    k = np.unique(label)
    v = np.asarray(range(len(k)))

    mapping_ar = np.zeros(k.max() + 1, dtype=dtype)
    mapping_ar[k] = v

    return mapping_ar[label]


def deep_update(d, u):
    """an dict update function that works with nested dicts"""
    import collections.abc

    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = deep_update(d.get(k, {}), v)
        else:
            d[k] = v
    return d

def remove_dictkey(d, k):
    """recursively remove a key from a nested dict"""
    import collections.abc
    if not isinstance(d, collections.abc.Mapping):
        return
    if k in d:
        del d[k]
    for sub in d.values():
        remove_dictkey(sub, k)



from __future__ import annotations

from functools import partial

import jax
import jax.numpy as jnp
from xtrain import unpack_x_y_sample_weight

from ..typing import *

def get_image_shape(batch):
    inputs, _, _ = unpack_x_y_sample_weight(batch)
    img = inputs['image']
    if img.ndim == 3:
        return (1,) + img.shape[:2]
    else:
        return img.shape[:3]

def generalized_binary_loss(
    logits: ArrayLike, gt: ArrayLike, gamma: int | float, beta: int | float
) -> Array:
    pred = jax.nn.softmax(logits)

    p_t = pred * gt + (1 - pred) * (1 - gt)
    ff = 1 - p_t

    ce = optax.softmax_cross_entropy(logits, gt)

    loss = (ff**gamma) * (ce**beta)

    return loss


binary_focal_factor_loss = partial(generalized_binary_loss, beta=0)

binary_focal_crossentropy = partial(generalized_binary_loss, beta=1)

binary_crossentropy = partial(generalized_binary_loss, gamma=0, beta=1)


def sum_over_boolean_mask(loss: ArrayLike, mask: ArrayLike) -> Array:
    mask = mask.reshape(mask.shape[0], 1)

    loss = loss.reshape(loss.shape[0], -1)
    loss = loss.mean(axis=1, keepdims=True).sum(where=mask)

    return loss


def mean_over_boolean_mask(loss: ArrayLike, mask: ArrayLike) -> Array:
    mask = mask.reshape(mask.shape[0], 1)
    n_instances = jnp.count_nonzero(mask) + 1e-8

    loss = loss.reshape(loss.shape[0], -1)
    loss = loss.mean(axis=1, keepdims=True).sum(where=mask)
    loss /= n_instances

    return loss



from __future__ import annotations

from functools import partial
from typing import Optional, Sequence, Callable

import flax.linen as nn
import jax.numpy as jnp
from flax.linen import normalization
from flax.linen import activation

from ..typing import DataDict, Array, ArrayLike, Any
from .common import DropPath, DefaultUnpicklerMixin, ChannelAttention, GRN

""" Implements the convnext encoder. Described in https://arxiv.org/abs/2201.03545
Original implementation: https://github.com/facebookresearch/ConvNeXt
"""
_LayerNorm = partial(nn.LayerNorm, epsilon=1e-6)

class _Block(nn.Module):
    """ConvNeXt Block.
    Args:
        drop_path (float): Stochastic depth rate. Default: 0.0
        layer_scale_init_value (float): Init value for Layer Scale. Default: 1e-6.
    """

    drop_rate: float = 0.4
    kernel_size: int = 7
    se_ratio: int = 16
    normalization: Callable=normalization.GroupNorm
    activation: Callable[[ArrayLike], Array]=nn.gelu
    version: str = "1"
    dtype: Any = None

    @nn.compact
    def __call__(self, x: ArrayLike, *, deterministic: Optional[bool] = True) -> Array:
        dim = jnp.asarray(x).shape[-1]
        ks = self.kernel_size

        shortcut = x

        x = nn.Conv(dim, (ks, ks), feature_group_count=dim, dtype=self.dtype)(x)
        x = self.normalization(dtype=self.dtype)(x)
        x = nn.Dense(dim * 4, dtype=self.dtype)(x)
        x = self.activation(x)
        if self.version == "2":
            x = GRN()(x)
        x = nn.Dense(dim, dtype=self.dtype)(x)

        if self.se_ratio > 0:
            x = ChannelAttention(self.se_ratio, dtype=self.dtype)(x)

        x = DropPath(self.drop_rate)(x, deterministic=deterministic)

        x = x + shortcut

        return x


class ConvNeXt(nn.Module, DefaultUnpicklerMixin):
    """ConvNeXt CNN backbone

    Attributes:
        patch_size: Stem patch size
        depths: Number of blocks at each stage.
        dims: Feature dimension at each stage.
        drop_path_rate: Stochastic depth rate.
        layer_scale_init_value: Init value for Layer Scale.
        out_channels:
            FPN output channels. Setting this to -1 disable the FPN, in which case
            the model output only encoder outputs.
    """

    patch_size: int = 4
    depths: Sequence[int] = (3, 3, 9, 3)
    dims: Sequence[int] = (96, 192, 384, 768)
    drop_path_rate: float = 0.4
    kernel_size: int = 7
    normalization: Callable=nn.GroupNorm
    se_ratio: int = 0
    activation: Callable[[ArrayLike], Array]=nn.gelu
    deterministic: bool|None = None
    version: str = "1"
    dtype: Any = None

    @nn.compact
    def __call__(
        self, x: ArrayLike, *, deterministic: Optional[bool] = None,
    ) -> Sequence[Array]:
        """
        Args:
            x: Image input.
            training: Whether run the network in training mode (i.e. with stochastic depth)

        Returns:
            A list of feture arrays at different scaling, from higest to lowest resolution.
        """
        if deterministic is None:
            deterministic = self.deterministic
        dp_rate = 0
        outputs = []
        for k in range(len(self.depths)):
            if k == 0:
                ps = self.patch_size
                x = nn.Conv(self.dims[k], (ps, ps), strides=(ps, ps))(x)
                x = self.normalization()(x)
            else:
                x = self.normalization()(x)
                x = nn.Conv(self.dims[k], (2, 2), strides=(2, 2))(x)

            for _ in range(self.depths[k]):
                x = _Block(
                    dp_rate,
                    kernel_size=self.kernel_size,
                    se_ratio=self.se_ratio,
                    normalization=self.normalization,
                    activation=self.activation,
                    version=self.version,
                    dtype=self.dtype,
                )(x, deterministic=deterministic)
                dp_rate += self.drop_path_rate / (sum(self.depths) - 1)

            outputs.append(x)

        return outputs

    @classmethod
    def get_preconfigured(cls, model_type:str, patch_size=4, **kwargs):
        if model_type == "tiny":
            return cls(patch_size, depths=(3, 3, 9, 3), dims=(96, 192, 384, 768), se_ratio=16, **kwargs)
        elif model_type == "small":
            return cls(patch_size, depths=(3, 3, 27, 3), dims=(96, 192, 384, 768), se_ratio=16, **kwargs)
        elif model_type == "base":
            return cls(patch_size, depths=(3, 3, 27, 3), dims=(128, 256, 512, 1024), se_ratio=16, **kwargs)
        elif model_type == "large":
            return cls(patch_size, depths=(3, 3, 27, 3), dims=(192, 384, 768, 1536), se_ratio=16, **kwargs)
        elif model_type == "x-large":
            return cls(patch_size, depths=(3, 3, 27, 3), dims=(256, 512, 1024, 2048), se_ratio=16, **kwargs)
        elif model_type == "tiny_v2":
            return cls(
                patch_size, 
                depths=(3, 3, 9, 3), 
                dims=(96, 192, 384, 768),
                se_ratio=-1,
                normalization=_LayerNorm,
                version="2",
                **kwargs,
            )
        elif model_type == "base_v2":
            return cls(
                patch_size, 
                depths=(3, 3, 27, 3), 
                dims=(128, 256, 512, 1024),
                se_ratio=-1,
                normalization=_LayerNorm,
                version="2",
                **kwargs,
            )
        elif model_type == "large_v2":
            return cls(
                patch_size, 
                depths=(3, 3, 27, 3), 
                dims=(192, 384, 768, 1536),
                se_ratio=-1,
                normalization=_LayerNorm,
                version="2",
                **kwargs,
            )
        elif model_type == "huge_v2":
            return cls(
                patch_size, 
                depths=(3, 3, 27, 3), 
                dims=(352, 704, 1408, 2816),
                se_ratio=-1,
                normalization=_LayerNorm,
                version="2",
                **kwargs,
            )            
        else:
            raise ValueError(f"unknown model type {model_type}")



""" 
Attributes:
    model_urls: URLs for build-in pretrain models. e.g model_urls["default"].
"""
from __future__ import annotations

import logging
import math
import time
import os

from functools import reduce, partial
from typing import Mapping, Sequence, Tuple, Any

import cv2
import flax.linen as nn
import jax
import jax.numpy as jnp
import numpy as np

from ..ops import patches_to_label, bboxes_of_patches, crop_and_resize_patches
from ..utils import load_from_pretrained
from ..typing import Array, ArrayLike

Shape = Sequence[int]

logger = logging.getLogger(__name__)

def _remove_edge_instances(pred, pos, patch_sz, img_sz):
    bboxes = pred['bboxes']
    removal = np.zeros([bboxes.shape[0]], dtype=bool)

    dim = len(img_sz)

    removal |= ((bboxes[:, :dim] <= 0) & (pos > 0)).any(axis=-1)
    removal |= ((bboxes[:, dim:] >= patch_sz) & (pos + patch_sz < img_sz)).any(axis=-1)

    pred["segmentation_is_valid"] &= ~removal


def _to_polygons(
    predictions, mask,
    *,
    segmentation_threshold=0.0,
    chain_approx=cv2.CHAIN_APPROX_SIMPLE,
) -> Tuple[Sequence[Any], np.ndarray]:
    polygons = []

    y0s = np.asarray(predictions["segmentation_y0_coord"])
    x0s = np.asarray(predictions["segmentation_x0_coord"])
    segs = np.asarray(predictions["segmentations"] >= segmentation_threshold).astype("uint8")
    mask = np.array(mask) 

    segs = segs.squeeze(1)

    for k in range(segs.shape[0]):
        if mask[k]:
            c, _ = cv2.findContours(segs[k], cv2.RETR_EXTERNAL, chain_approx)
            max_len_element = reduce(
                lambda a, b: a if len(a) >= len(b) else b,
                c,
                np.zeros([0, 1, 2], dtype=int),
            )
            polygon = max_len_element.squeeze(1).astype(float)
            if len(polygon) > 0:
                polygons.append(polygon + [x0s[k], y0s[k]] + .5)
            else:
                mask[k] = False

    return polygons, mask


def _to_mesh(
    predictions, mask,
    *,
    segmentation_threshold=0.0,
    step_size=1,
):
    from skimage.measure import marching_cubes
    meshes = []

    z0s = np.asarray(predictions["segmentation_z0_coord"])
    y0s = np.asarray(predictions["segmentation_y0_coord"])
    x0s = np.asarray(predictions["segmentation_x0_coord"])
    segs = np.asarray(predictions["segmentations"])
    mask = np.array(mask) 

    for k in range(segs.shape[0]):
        if mask[k]:
            try:
                # verts, faces, norms, _ = marching_cubes(seg, allow_degenerate=False)
                verts, faces, _, _ = marching_cubes(segs[k], step_size=step_size, level=segmentation_threshold)
                verts += [z0s[k], y0s[k], x0s[k]]

                meshes.append(dict(verts=verts, faces=faces))

            except:
                mask[k] = False

    return meshes, mask  


def _clean_up_mesh(mesh, n=500, regularization=0.1):
    from vedo import Mesh
    
    vedo_mesh = (
        Mesh([mesh["verts"], mesh["faces"]])
        .decimate(n=n, regularization=regularization)
    )
    return dict(
        verts=vedo_mesh.vertices,
        faces=np.array(vedo_mesh.cells),
    )


def _nms(boxes, iou_threshold, selected, asort=None):
    from lacss.ops import box_iou_similarity

    if asort is not None:
        boxes = boxes[asort]
        selected = selected[asort]

    sm = np.array(box_iou_similarity(boxes, boxes))
    sm = np.triu(sm, k=1)

    assert len(selected) == boxes.shape[0]

    for k in range(len(selected)):
        if selected[k]:
            selected &= sm[k] < iou_threshold
    
    # inverse selected` to the original order
    if asort is not None:
        selected = selected[asort.argsort()]

    return selected


def _predict_partial(model, params, image):
    model = model.copy()
    model.segmentor = None
    model.segmentor_3d = None

    lpn_out, seg_features = [], []
    for p in params:
        model_out =  model.apply(dict(params=p), image)
        lpn_out.append( model_out['detector'] )
        seg_features.append(model_out['seg_features'])

    lpn_out = jax.tree_util.tree_map(
        lambda *x: jnp.stack(x), * lpn_out
    )
    lpn_out = dict(
        logits = lpn_out['logits'].mean(axis=0),
        regressions =  lpn_out['regressions'].mean(axis=0),
        ref_locs = lpn_out['ref_locs'][0],
    )
    lpn_out['pred_locs'] = lpn_out['ref_locs'] + lpn_out['regressions'] * model.detector.feature_scale

    return seg_features, lpn_out


@partial(jax.jit, static_argnums=0)
def _model_fn(model, params, image):
    from lacss.modules.lpn import generate_predictions

    seg_features, lpn_out = _predict_partial(model, params, image)

    predictions = generate_predictions(model.detector, lpn_out)

    assert len(seg_features) == len(params)

    seg_predictions = []
    if image.ndim == 3 and model.segmentor is not None:
        for x, p in zip(seg_features, params):
            seg_predictions.append( 
                model.segmentor.apply(
                    dict(params=p['segmentor']), 
                    x, 
                    predictions['locations']
                )['predictions'] 
            )

    elif image.ndim == 4 and model.segmentor_3d is not None:
        seg_predictions = []
        for x, p in zip(seg_features, params):
            seg_predictions.append( 
                model.segmentor_3d.apply(
                    dict(params=p['segmentor_3d']), 
                    x, 
                    predictions['locations']
                )['predictions'] 
            )

    if len(seg_predictions) > 0:
        predictions['segmentations'] = jnp.mean(
            jnp.array([ x['segmentations'] for x in seg_predictions]),
            axis = 0,
        )

        seg_prediction = seg_predictions[0]
        del seg_prediction['segmentations']

        predictions.update( seg_prediction )
        
    return predictions


class Predictor:
    """Main class interface for model deployment. This is the only class you
    need if you don't train your own model

    Examples:
        The most common use case is to use a build-in pretrained model.

            import lacss.deploy

            # look up the url of a build-in mode
            url = lacss.deploy.model_urls["default"]

            # create the predictor instance
            predictor = lacss.deploy.Predictor(url)

            # make a prediction
            label = predictor.predict(image)

    Attributes:
        module: The underlying FLAX module
        params: Model weights.
    """
    def __init__(
            self, 
            url: str | os.PathLike | tuple[nn.Module, dict],
            *,
            f16 = False,
            grid_size:int = 1088,
            step_size:int = 1024,
            grid_size_3d:int = 384,
            step_size_3d:int = 320,
            mask_size:int = 36,
            mc_step_size:int = 2,
        ):
        """Construct Predictor

        Args:
            url: A URL or local path to the saved model.
                URLs for build-in pretrained models can be found in lacss.deploy.model_urls
        """

        if isinstance(url, tuple) and len(url) == 2:
            if not isinstance(url[0], nn.Module):
                raise ValueError(
                    "Initiaize the Predictor with a tuple, but the first element is not a Module."
                )

            module, params = url

        else:
            module, params = load_from_pretrained(url)


        if isinstance(params, Mapping):
            params = [params]

        self.module = module
        self.params = params

        self.f16 = f16
        if self.f16:
            self.params = jax.tree_util.tree_map(
                lambda x: x.astype("float16"), self.params,
            )

        assert step_size < grid_size, f"step_size ({step_size}) not smaller than grid_size ({grid_size})"
        assert grid_size % 32 == 0, f"grid_size ({grid_size}) is not divisable by 32"
        assert step_size_3d < grid_size_3d, f"step_size ({step_size_3d}) not smaller than grid_size ({grid_size_3d})"
        assert grid_size_3d % 32 == 0, f"grid_size ({grid_size_3d}) is not divisable by 32"

        self.gs = grid_size
        self.ss = step_size

        self.gs_3d = grid_size_3d
        self.ss_3d = step_size_3d

        self.mask_size = mask_size
        self.mc_step_size = mc_step_size


    def _format_image(self, image, target_shape, normalize=True):
        from skimage.transform import resize

        # process the data in float32 to avoid overflow
        # only convert to float16 (if needed) at the end

        image = image.astype("float32")

        orig_shape = image.shape[:-1]        
        
        if normalize:
            image = image - image.mean()
            image = image / (image.std() + 1e-6)

        if target_shape is None: 
            target_shape = np.array(orig_shape)

        else:
            target_shape = np.broadcast_to(target_shape, [len(orig_shape)])

            scaling = target_shape / orig_shape

            if np.all(np.abs(scaling - 1) < 0.1): #ignore small scaling
                target_shape = np.array(orig_shape)

            else:
                image = resize(image, target_shape)

        if self.f16:
            image = image.astype("float16")

        logger.debug(f"resized image data from {orig_shape} to {target_shape}")

        return image, target_shape


    def _call_model(self, image, threshold, to_cpu=True):
        logger.debug(f"process image patch of {image.shape}")

        # pad image to fixed input sizes
        img_sz = image.shape[:-1]
        padding_shape = [(x - 1) // 32 * 32 + 32 for x in img_sz]

        if image.ndim == 3: #2D input
            if max(padding_shape) <=  544:
                padding_shape = [544, 544]

            elif max(padding_shape) <=  1088:
                padding_shape = [1088, 1088]

        else:
            # for 3D we require isotropic dimension
            padding_shape = [max(padding_shape)] * 3

            if max(padding_shape) <= 256:
                padding_shape = [256, 256, 256]

            elif max(padding_shape) <=  384:
                padding_shape = [384, 384, 384]

        padding = [ [0, s - s0] for s, s0 in zip(padding_shape, image.shape[:-1])]
        padding += [[0, 0]]

        image = np.pad(image, padding)

        # make 3-ch
        if image.shape[-1] == 1:
            image = np.repeat(image, 3, axis=-1)
        elif image.shape[-1] == 2:
            image = np.stack([image, np.zeros_like(image[..., :1])], axis=-1)

        logger.debug(f"call model with data {image.shape}")
        predictions = _model_fn(self.module, self.params, image)

        predictions["bboxes"] = bboxes_of_patches(
            predictions,
            threshold=threshold,
            image_shape=img_sz,
        )

        if to_cpu:
            predictions = jax.tree_util.tree_map(
                lambda x: np.array(x),
                predictions,
            )

        return predictions


    def _process_image(self, image, threshold):
        img_sz = image.shape[:-1]
        img_dim = len(img_sz)

        # no grid for small images
        if (
            (img_dim == 2 and max(img_sz) <= self.gs) or
            (img_dim == 3 and max(img_sz) <= self.gs_3d)
         ) :
            return self._call_model(image, threshold), True

        
        gs, ss = (self.gs, self.ss) if img_dim == 2 else (self.gs_3d, self.ss_3d)

        grid_positions = [slice(0, max(d-(gs-ss), 1), ss) for d in img_sz]
        grid_positions = np.moveaxis(np.mgrid[grid_positions], 0, -1)
        grid_positions = grid_positions.reshape(-1, img_dim)

        predictions = []
        for pos in grid_positions:
            slices = (slice(x, x + gs) for x in pos)
            patch = image.__getitem__(tuple(slices))
            patch_sz = patch.shape[:-1]
            
            patch_pred = self._call_model(patch, threshold)
            _remove_edge_instances(patch_pred, pos, patch_sz, img_sz)

            n_det = np.count_nonzero(patch_pred["segmentation_is_valid"])
            if n_det > 0:
                z0, y0, x0 = pos if img_dim == 3 else (0, pos[0], pos[1])
                patch_pred["bboxes"] += np.r_[pos, pos]
                patch_pred["segmentation_z0_coord"] += z0
                patch_pred["segmentation_y0_coord"] += y0
                patch_pred["segmentation_x0_coord"] += x0
                predictions.append(patch_pred)
            
        preds = jax.tree_util.tree_map(lambda *x: np.concatenate(x), *predictions)

        return preds, False


    def _filter_predictions(self, predictions, score_threshold, segmentation_threshold, min_area, nms_iou, is_sorted):
        instance_mask = predictions["segmentation_is_valid"]

        instance_mask &= predictions["scores"] >= score_threshold

        if min_area > 0:
            areas = np.count_nonzero(
                predictions["segmentations"] > segmentation_threshold, 
                axis=(1, 2, 3),
            )
    
            instance_mask &= areas >= min_area

        if nms_iou > 0 and nms_iou < 1:
            instance_mask = _nms(
                predictions["bboxes"], 
                nms_iou,
                instance_mask,
                predictions["scores"].argsort()[::-1] if not is_sorted else None,
            )

        return instance_mask


    def predict(
        self,
        image: ArrayLike,
        *,
        output_type: str = "label",
        reshape_to: int|tuple[int]|np.ndarray|None = None,
        min_area: float = 0,
        score_threshold: float = 0.5,
        segmentation_threshold: float = 0.5,
        nms_iou: float = 1,
        normalize: bool = True,
        remove_out_of_bound: bool|None = None,
    ) -> dict:
        """Predict segmentation.

        Args:
            image: A ndarray of (h,w,c) or (d,h,w,c) format. c must be 1-3

        Keyword Args:
            output_type: "label" | "contour" | "bbox"
            reshape_to: If not None, the input image will be resized internally before send
                to the model. The results will be resized back to the scale of the orginal input image.
            min_area: Minimum area of a valid prediction.
            score_threshold: Min score needed to be included in the output.
            segmentation_threshold: Threshold value for segmentation
            nms_iou: IOU threshold value for non-max-supression post-processing

        Returns:
            For "label" output:

                - pred_scores: The prediction scores of each instance.
                - pred_label: a 2D image label. 0 is background.

            For "contour" output:

                - pred_scores: The prediction scores of each instance.
                - pred_contours: a list of polygon arrays in x-y format.

            For "bbox" output (ie MaskRCNN):

                - pred_scores: The prediction scores of each instance.
                - pred_bboxes: The bounding-boxes of detected instances in y0x0y1x1 format
                - pred_masks:  A 3d array representing (rescaled) segmentation mask within bboxes
        """
        image = np.asarray(image)
        
        logger.debug(f"started prediction with image {image.shape}")
        start_time = time.time()

        if remove_out_of_bound is not None:
            import warnings
            warnings.warn("remove_out_of_bound is deprecated", DeprecationWarning, 2)

        if image.ndim == 2 or image.shape[-1] > 3:
            logger.warning("input seems to have no channel dim. Add one")
            image = image[..., None]

        assert image.ndim == 3 or image.ndim == 4, f"illegal image dim {image.shape}"

        assert image.shape[-1] <= 3, f"input image has more than 3 channels"

        if not output_type in ("bbox", "label", "contour", "_raw"):
            raise ValueError(
                f"output_type should be 'bbox'|'label'|'contour'. Got {output_type} instead."
            )

        img_shape = image.shape[:-1]

        is_3d = image.ndim == 4

        image, reshape_to = self._format_image(image, reshape_to, normalize=normalize)

        scaling = reshape_to / img_shape

        seg_logit_threshold = math.log(segmentation_threshold / (1 - segmentation_threshold))

        logger.debug(f"done preprocessing")

        preds, is_sorted = self._process_image(image, seg_logit_threshold)

        logger.debug(f"filter detections...")
        instance_mask = self._filter_predictions(
            predictions=preds,
            score_threshold=score_threshold, 
            segmentation_threshold=segmentation_threshold,
            min_area=min_area * np.prod(scaling),
            nms_iou=nms_iou,
            is_sorted=is_sorted,
        )
        logger.debug(f"filter detections...done")

        # generate outputs
        if output_type == "_raw":
            preds["segmentation_is_valid"] = instance_mask
            return preds

        elif output_type == "bbox":
            target_shape = np.broadcast_to([self.mask_size], [len(img_shape)])
            segmentations = crop_and_resize_patches(
                preds, preds["bboxes"],
                target_shape=tuple(target_shape),
                convert_logits=True,
            )
            results = dict(
                pred_scores=preds["scores"][instance_mask],
                pred_bboxes=preds["bboxes"][instance_mask] / np.r_[scaling, scaling],
                pred_masks=np.asarray(segmentations)[instance_mask],
                pred_locations=preds["locations"][instance_mask] / scaling,
            )

        elif output_type == "contour":
            if not is_3d:
                polygons, instance_mask = _to_polygons(
                    preds, instance_mask, 
                    segmentation_threshold=seg_logit_threshold,
                )
                polygons = [c / scaling[::-1] for c in polygons]

                results = dict(
                    pred_scores=preds["scores"][instance_mask],
                    pred_bboxes=preds["bboxes"][instance_mask] / np.r_[scaling, scaling],
                    pred_contours=polygons,
                )
            else:
                meshes, instance_mask = _to_mesh(
                    preds, instance_mask,
                    segmentation_threshold=seg_logit_threshold,
                    step_size=self.mc_step_size,
                )

                for mesh in meshes:
                    mesh['verts'] /=  scaling

                # clean up
                meshes = [_clean_up_mesh(m) for m in meshes]

                results = dict(
                    pred_scores=preds["scores"][instance_mask],
                    pred_bboxes=preds["bboxes"][instance_mask] / np.r_[scaling, scaling],
                    pred_contours=meshes,
                )                

        else:  # Label
            label = patches_to_label(
                preds, 
                tuple(reshape_to),
                mask=instance_mask,
                score_threshold=0,
                threshold=seg_logit_threshold,
            )
            if not (scaling == 1.0).all():
                label = jax.image.resize(
                    label, img_shape, "nearest",
                )

            results = dict(
                pred_scores=preds["scores"][instance_mask],
                pred_label=np.asarray(label),
            )

        elapsed = (time.time() - start_time) * 1000
        logger.debug(f"done prediction in {elapsed:.2f} ms")

        return results




"""Vision Transformer.
Based on scenic implementation: https://github.com/google-research/scenic/
"""

from typing import Any, Callable, Optional, Sequence

import flax.linen as nn
import jax
import jax.numpy as jnp
import numpy as np

from .common import DropPath, IdentityLayer, FFN, PositionEmbedding1D, PositionEmbedding2D
from ..typing import Array, ArrayLike

Initializer = Callable[[jnp.ndarray, Sequence[int], jnp.dtype], jnp.ndarray]

class MAPHead(nn.Module):
    """Multihead Attention Pooling."""
    mlp_dim: Optional[int] = None  # Defaults to 4x input dim
    num_heads: int = 12

    @nn.compact
    def __call__(self, x):
        x_shape = x.shape
        x = x.reshape(-1, x_shape[-3] * x_shape[-2], x_shape[-1])
        probe = self.param(
            'probe', 
            nn.initializers.xavier_uniform(), 
            (1, 1, x.shape[-1]),
            self.param_dtype,
        )
        probe = probe.astype(x.dtype)
        probe = jnp.tile(probe, [x.shape[0], 1, 1])

        x = nn.MultiHeadDotProductAttention(
            num_heads=self.num_heads, 
            kernel_init=nn.initializers.xavier_uniform(),
        )(probe, x)

        y = nn.LayerNorm()(x)
        x = x + FFN(
            self.mlp_dim,
            dropout_rate=0.0,
        )(y, deterministic=True)

        return x[:, 0].reshape(*x_shape[:-3], -1)


class VitEncoder(nn.Module):
    """Transformer Encoder.

    Attributes:
      n_layers: Number of layers.
      n_heads: The number of heads for multi-head self-attention.
      dim_ffn: Dimension of the mlp on top of attention block.
      pos_embedding_type: The type of positional embeddings to add to the
        input tokens.
      dropout: Dropout rate.
      stochastic_depth: probability of dropping a layer linearly grows
        from 0 to the provided value. Our implementation of stochastic depth
        follows timm library, which does per-example layer dropping and uses
        independent dropping patterns for each skip-connection.
    """
    n_layers: int = 6
    n_heads: int = 12
    dim_ffn: int|None = None
    pos_embedding_type: str = 'learned_1d'
    dropout: float = 0.1
    attention_dropout: float = 0.1
    stochastic_depth: float = 0.0

    @nn.compact
    def __call__(self, inputs: ArrayLike, cls_token: ArrayLike|None=None, *, training: bool = False):
        """Applies Transformer model on the inputs.

        Args:
          inputs: Input tokens of shape [..., num_tokens, channels].
          cls_token: addtional class token if not None
          training: If in training mode, dropout and stochastic depth is applied.

        Returns:
          Encoded tokens and cls_token
        """

        input_shape = inputs.shape
        H, W, C = input_shape[-3:]

        x = inputs.reshape((-1,) + (H*W, C))

        if self.pos_embedding_type == 'learned_1d':
            x = x + PositionEmbedding1D()(x)

        elif self.pos_embedding_type == 'sinusoidal_2d':
            pos = PositionEmbedding2D(
                posemb_init=None,
            )(input_shape)
            x = x + pos.reshape(-1, C)

        elif self.pos_embedding_type == 'learned_2d':
            pos = PositionEmbedding2D()(input_shape)
            x = x + pos.reshape(-1, C)
        
        elif self.pos_embedding_type != 'none':
            raise ValueError("invalid pos_embedding_type {self.pos_embedding_type}")

        if cls_token:
            assert cls_token.shape == input_shape[:-3] + (C,), f"invalid class token shape {cls_token.shape}"
            x = jnp.c_[x, cls_token.reshape(-1, 1, C)]

        # x = nn.Dropout(rate=self.dropout)(x, deterministic=not training)

        for lyr in range(self.n_layers):
            shortcut = x
            x = nn.Layernorm()(x)
            x = nn.MultiHeadAttention(
                num_heads=self.n_heads,
                dropout_rate=self.attention_dropout,
                kernel_init=nn.initializers.xavier_uniform(),
                broadcast_dropout=False,
                name=f"attention_{lyr}",
            )(x, deterministic=not training)
            
            x = FFN(dim=self.dim_ffn, dropout_rate=self.dropout, name="ffn_{lyr}")(
                x, deterministic=not training
            )

            droppath_rate = lyr/self.n_layers*self.stochastic_depth
            x = DropPath(droppath_rate)(x, deterministic=not training)

            x = shortcut + x

        if cls_token:
            cls_token = x[..., -1, :].reshape(input_shape[:-3] + (C,))
            x = x[..., :-1, :].reshape(input_shape)
            return x, cls_token
        else:            
            x = x.reshape(input_shape)
            return x, None


class ViT(nn.Module):
    """Vision Transformer model.

      Attributes:
        dim: Size of the hidden state of the output of model's stem.
        dim_ffn: Dimension of the mlp on top of attention block.
        n_layers: Number of layers.
        n_heads: Number of self-attention heads.
        patch_size: specify the patch dimension for tokenization
        pos_embedding_type: The type of positional embeddings to add to the
            tokens at the beginning of the transformer encoder. 
        dropout: Dropout rate.
        attention_dropout: Dropout for attention heads.

        n_cls: Number of output classes.
        representation_size: Size of the representation layer in the model's head.
            if None, we skip the extra projection + tanh activation at the end.
        classifier: type of the classifier layer. Options are 'gap', 'gmp', 'gsp',
            'token', or None to return encoder features
    """
    # transformer parameters
    dim: int = 256
    dim_ffn: int|None = None
    n_layers: int = 8
    n_heads: int = 12
    patch_size: tuple[int,int] = (8, 8)
    pos_embedding_type: str = 'learned_1d'
    dropout: float = 0.1
    attention_dropout: float = 0.1
    stochastic_depth: float = 0.0

    # classifier parameters
    n_cls: int = -1
    representation_size: Optional[int] = None
    classifier_type: str|None = None

    @nn.compact
    def __call__(self, x: jnp.ndarray, *, training: bool=False):
        """ViT model

        Args:
          inputs: Input [..., H, W, C].
          training: If in training mode, dropout and stochastic depth is applied.

        Returns:
          features: [..., H//patch_size, W//patch_size, n_dim_hidden]
        """
        input_shape = x.shape

        x = x.reshape((-1,) + input_shape[-3:]) # [b, h, w, c]

        fh, fw = self.patch_size
        x = nn.Conv(
            self.dim, (fh, fw),
            strides=(fh, fw),
            padding='VALID',
            name='embedding',
        )(x)

        output_shape = x.shape

        # x = jnp.reshape(x.shape[0], -1, x.shape[-1]) # merge H and W

        # Add a class token if suitable
        if self.classifier_type == 'token':
            cls_token = self.param('cls', nn.initializers.zeros, (1, x.shape[-1]), self.param_dtype)
            cls_toekn = cls_token.astype(x.dtype)
            cls_token = jnp.tile(cls_token, [x.shape[0], 1])
        else:
            cls_token = None

        x, cls_token = VitEncoder(
            dim_ffn=self.dim_ffn,
            n_layers=self.n_layers,
            n_heads=self.n_heads,
            pos_embedding_type=self.pos_embedding_type,
            dropout=self.dropout,
            attention_dropout=self.attention_dropout,
            stochastic_depth=self.stochastic_depth,
            name='Transformer',
        )(x, cls_token, training=training)

        if self.classifier_type is None:
            return x.reshape(output_shape)

        if self.classifier_type in ('token', '0'):
            x = cls_token

        elif self.classifier_type in ('gap', 'gmp', 'gsp'):
            fn = {'gap': jnp.mean, 'gmp': jnp.max,
                  'gsp': jnp.sum}[self.classifier_type]
            x = fn(x, axis=1)

        elif self.classifier_type == 'map':
            x = MAPHead(
                num_heads=self.n_heads, 
                mlp_dim=self.dim_ffn, 
            )(x)

        else:
            raise ValueError(f'Unknown classifier {self.classifier_type}')

        if self.representation_size is not None:
            x = nn.Dense(self.representation_size, name='pre_logits')(x)
            x = nn.tanh(x)
        else:
            x = IdentityLayer(name='pre_logits')(x)

        # If self.n_cls <= 0, we just return the representations
        if self.n_cls > 0:
            x = nn.Dense(
                self.n_cls,
                kernel_init=nn.initializers.zeros,
                name='output_projection',
            )(x)

        output = x.reshape(input_shape[:-3] + x.shape[-1:])

        return output



from __future__ import annotations

import flax.linen as nn
import jax
from .common import FFN, PositionEmbedding1D
from ..typing import Any, Sequence, Array, ArrayLike
jnp = jax.numpy

class VideoIntegrator(nn.Module):
    n_heads: int = 4
    dropout: float = 0.2
    ff_dropout: float = 0.2
    dtype: Any = None

    @nn.compact
    def __call__(self, feature:Sequence[ArrayLike], video_refs:tuple[Sequence[ArrayLike], ArrayLike|None], deterministic:bool=True)->Sequence[Array]:
        ref_feature, ref_mask = video_refs

        out = []
        for k in range(4):
            x = jnp.asarray(feature[k])[..., None, :]

            y = jnp.moveaxis(ref_feature[k], 0, -2)
            pos_embedding = PositionEmbedding1D()(y)

            x = nn.LayerNorm(epsilon=1e-6, dtype=self.dtype)(x)
            x = nn.MultiHeadDotProductAttention(
                self.n_heads,
                dropout_rate=self.dropout,
                dtype=self.dtype,
            )(
                inputs_q = x,
                inputs_k = y + pos_embedding,
                inputs_v = y,
                deterministic=deterministic,
                sow_weights=True,
                mask=ref_mask,
            )

            x = FFN(dropout_rate=self.ff_dropout, dtype=self.dtype)(x, deterministic=deterministic)

            x = x.squeeze(-2)

            out.append(feature[k] + x)

        return out



""" Ops on bounding-boxes

All functions here are degisned to work as either a numpy op or a jax op depending on
the data type of the input.
"""
from __future__ import annotations

import jax
import jax.numpy as jnp
import numpy as np

from ..typing import *


def box_area(box: ArrayLike) -> ArrayLike:
    """Computes area of boxes.
    Args:
      box: a float Tensor with [..., N, 2d].

    Returns:
      a float Tensor with [..., N]
    """
    ndim = box.shape[-1] // 2
    assert ndim * 2 == box.shape[-1]

    min_vals = box[..., :ndim]
    max_vals = box[..., ndim:]

    return (max_vals - min_vals).prod(axis=-1)


def box_intersection(gt_boxes: ArrayLike, boxes: ArrayLike) -> ArrayLike:
    """Compute pairwise intersection areas between boxes.

    Args:
      gt_boxes: [..., N, 2d]
      boxes: [..., M, 2d]

    Returns:
      a float Tensor with shape [..., N, M] representing pairwise intersections.
    """
    if isinstance(gt_boxes, jnp.ndarray) & isinstance(boxes, jnp.ndarray):
        minimum = jnp.minimum
        maximum = jnp.maximum
    else:
        minimum = np.minimum
        maximum = np.maximum

    ndim = gt_boxes.shape[-1] // 2
    assert ndim * 2 == gt_boxes.shape[-1]
    assert ndim * 2 == boxes.shape[-1]

    min_vals_1 = gt_boxes[..., None, :ndim] # [..., N, 1, d]
    max_vals_1 = gt_boxes[..., None, ndim:]
    min_vals_2 = boxes[..., None, :, :ndim] # [..., 1, M, d]
    max_vals_2 = boxes[..., None, :, ndim:]

    min_max = minimum(max_vals_1, max_vals_2) #[..., N, M, d]
    max_min = maximum(min_vals_1, min_vals_2)

    intersects = maximum(0, min_max - max_min) # [..., N, M, d]

    return intersects.prod(axis=-1)


def box_iou_similarity(gt_boxes: ArrayLike, boxes: ArrayLike) -> ArrayLike:
    """Computes pairwise intersection-over-union between box collections.

    Args:
      gt_boxes: a float Tensor with [..., N, 2d].
      boxes: a float Tensor with [..., M, 2d].

    Returns:
      a Tensor with shape [..., N, M] representing pairwise iou scores.
    """
    intersections = box_intersection(gt_boxes, boxes)
    gt_boxes_areas = box_area(gt_boxes)
    boxes_areas = box_area(boxes)
    unions = gt_boxes_areas[..., None] + boxes_areas[..., None, :] - intersections

    ious = intersections / (unions + 1e-6)

    return ious

def yxhw_iou_similarity(yxhw_a, yxhw_b):
    """Computes pairwise IOU on bbox in yxhw format

    Args:
      gt_boxes: a float Tensor with [..., N, 2d].
      boxes: a float Tensor with [..., M, 2d].

    Returns:
      a Tensor with shape [..., N, M] representing pairwise iou scores.
    """
    def _yxhw2box(yxhw):
        dim = yxhw.shape[-1] // 2        
        center, span = yxhw[..., :dim], yxhw[..., dim:]
        return jnp.c_[center - span / 2, center + span / 2]
      
    return box_iou_similarity(_yxhw2box(yxhw_a), _yxhw2box(yxhw_b))

def iou_loss(gt_boxes: ArrayLike, boxes: ArrayLike) -> ArrayLike:
    """ IOU loss = 1 - IOU
    Args:
      gt_boxes: a float Tensor with [..., N, 2d].
      boxes: a float Tensor with [..., N, 2d].

    Returns:
      a Tensor with shape [..., N] representing pairwise loss.
    """

    loss = 1 - jax.vmap(box_iou_similarity, in_axes=-2, out_axes=-2)(
      gt_boxes[..., None, :, :], boxes[..., None, :, :],
    )
    return loss.squeeze(axis=-3)


def generalized_iou_loss(gt_boxes: ArrayLike, boxes: ArrayLike) -> ArrayLike:
    """ Loss_GIoU = 1 - IoU + |C - B union B_GT| / |C|
    where C is the smallest enclosing box for both B and B_GT. The resulting value has a gradient 
    for non-overlapping boxes. See  Zheng et al. [AAAI 2020] 

    Args:
      gt_boxes: a float Tensor with [..., N, 2d].
      boxes: a float Tensor with [..., N, 2d].

    Returns:
      a Tensor with shape [..., N] representing pairwise loss.
    """
    if isinstance(gt_boxes, jnp.ndarray) & isinstance(boxes, jnp.ndarray):
        minimum = jnp.minimum
        maximum = jnp.maximum
    else:
        minimum = np.minimum
        maximum = np.maximum

    ndim = gt_boxes.shape[-1] // 2
    assert ndim * 2 == gt_boxes.shape[-1]
    assert ndim * 2 == boxes.shape[-1]

    intersections = jax.vmap(box_intersection, in_axes=-2, out_axes=-2)(
      gt_boxes[..., None, :, :], boxes[..., None, :, :],
    ).squeeze(-3)
    gt_boxes_areas = box_area(gt_boxes)
    boxes_areas = box_area(boxes)
    unions = gt_boxes_areas + boxes_areas - intersections

    ious = intersections / (unions + 1e-6)

    mins = minimum(gt_boxes[..., :ndim], boxes[..., :ndim])
    maxs = maximum(gt_boxes[..., ndim:], boxes[..., ndim:])
    intersects = maximum(0, maxs - mins) # [..., N, d]

    C = intersects.prod(axis=-1)
    
    return 1 - ious + (C - unions) / (C + 1e-6)

def distance_iou_loss(gt_boxes: ArrayLike, boxes: ArrayLike) -> ArrayLike:
    """ Loss_distance_iou = 1 - IoU + \rho2(B, B_GT) / c2
    The correction term is "distance_sq between box center" / "distance_sq of enclosing box cornor"

    Args:
      gt_boxes: a float Tensor with [..., N, 2d].
      boxes: a float Tensor with [..., N, 2d].

    Returns:
      a Tensor with shape [..., N] representing pairwise loss.    
    """
    if isinstance(gt_boxes, jnp.ndarray) & isinstance(boxes, jnp.ndarray):
        minimum = jnp.minimum
        maximum = jnp.maximum
    else:
        minimum = np.minimum
        maximum = np.maximum

    ndim = gt_boxes.shape[-1] // 2
    assert ndim * 2 == gt_boxes.shape[-1]
    assert ndim * 2 == boxes.shape[-1]

    ious = jax.vmap(box_iou_similarity, in_axes=-2, out_axes=-2)(
      gt_boxes[..., None, :, :], boxes[..., None, :, :],
    ).squeeze(-3)
    mins = minimum(gt_boxes[..., :ndim], boxes[..., :ndim])
    maxs = maximum(gt_boxes[..., ndim:], boxes[..., ndim:])
    intersects = maximum(0, maxs - mins) # [..., N, d]
    c2 = (intersects ** 2).sum(axis=-1)

    center_gt = (gt_boxes[..., :ndim] + gt_boxes[..., ndim:]) / 2
    center = (boxes[..., :ndim] + boxes[..., ndim:]) / 2
    rho2 = ((center_gt - center) ** 2).sum(axis=-1)

    return 1 - ious + rho2/c2


def distance_similarity(pred_locations: ArrayLike, gt_locations: ArrayLike) -> Array:
    """Compute distance similarity matrix

        pairwise similarity = 1 / distance ^2

    Args:
        pred_locations: [N, d] use -1 to mask out invalid locations
        gt_locations: [K, d] use -1 to mask out invalid locations

    Returns:
        similarity_matrix: [N, k]
    """

    distances_matrix = pred_locations[:, None, :] - gt_locations
    distances_matrix = jnp.square(distances_matrix).sum(axis=-1)

    sm = 1.0 / (distances_matrix + 1e-8)

    # negative locations are invalid
    sm = jnp.where((pred_locations < 0).all(axis=-1)[:, None], 0, sm)
    sm = jnp.where((gt_locations < 0).all(axis=-1), 0, sm)

    return sm


def feature_matching(
    features_a: ArrayLike, features_b: ArrayLike, threshold: float,
    *,
    similarity_fn = None,
) -> tuple[Array, Array]:
    """Match predicted location to gt locations

    Args:
      features_a:r [N, d], points or bboxes
      features_b: [K, d], points or bboxes
      threshold: float. Min similarity score for match

    Keyword Args:
      similarity_fn: optional custom function for computing similarity score
          fn(feautes_a, feature_b) -> similarity matrix

    Returns:
      matches: [N], indices of the matches row in b 
      indicators: [N] bool
    """
    dim = features_a.shape[-1]
    if similarity_fn is None:
        similarity_fn = distance_similarity if dim <=3 else yxhw_iou_similarity

    sm = similarity_fn(features_a, features_b)
    matches = sm.argmax(axis=-1, keepdims=True)

    indicators = jnp.take_along_axis(sm, matches, -1) > threshold

    return matches.squeeze(-1), indicators.squeeze(-1)


def match_and_replace(gt_features, pred_features, threshold, *, similarity_fn = None):
    """replacing gt_locations with pred_locations if the close enough
    1. Each pred_location is matched to the closest gt_location
    2. For each gt_location, pick the matched pred_location with highest score
    3. if the picked pred_location is within threshold distance, replace the gt_location with the pred_location
    """

    n_gt_locs = gt_features.shape[0]
    n_pred_locs = pred_features.shape[0]

    matched_id, indicators = feature_matching(
        pred_features, 
        gt_features, 
        threshold,
        similarity_fn = similarity_fn 
    )
    matched_id = jnp.where(indicators, matched_id, -1)

    matching_matrix = (
        matched_id[None, :] == jnp.arange(n_gt_locs)[:, None]
    )  # true at matched gt(row)/pred(col), at most one true per col
    last_col = jnp.ones([n_gt_locs, 1], dtype=bool)
    matching_matrix = jnp.concatenate(
        [matching_matrix, last_col], axis=-1
    )  # last col is true

    # first true of every row
    matched_loc_ids = jnp.argmax(matching_matrix, axis=-1)

    training_locations = jnp.where(
        matched_loc_ids[:, None] == n_pred_locs,  # true: failed match
        gt_features,
        pred_features[matched_loc_ids] # TODO out-of-bound error on cpu
    )

    return training_locations



import threading

import grpc
import numpy as np

from . import proto

_AUTH_HEADER_KEY = "authorization"

def get_dtype(pixels:proto.Pixels) -> np.dtype:
    dt = np.dtype(pixels.dtype)

    if pixels.bindata.endianness == proto.BinData.Endianness.BIG:
        dt = dt.newbyteorder(">")
    else:
        dt = dt.newbyteorder("<")
    
    return dt


def decode_image(pixels:proto.Pixels) -> np.ndarray:
    if pixels.size_t > 1:
        raise ValueError("Image data has a non-singleton T dimension.")

    if pixels.size_c > 3:
        raise ValueError("Image data has more than 3 channels.")

    np_img = np.frombuffer(
        pixels.bindata.data, 
        dtype=get_dtype(pixels),
    ).astype("float32")

    # The dimension_order describe axis order but in the F_order convention
    # Numpy default is C_order, so we reverse the sequence. Lacss expect the 
    # final dimension order to be "ZYXC"
    dim_order_c = pixels.dimension_order[::-1].upper()
    dims = dict(
        Z = pixels.size_z or 1,
        Y = pixels.size_y or 1,
        X = pixels.size_x or 1,
        C = pixels.size_c or 1,
        T = 1,
    )
    dim_orig = [dim_order_c.find(k) for k in "ZYXCT"]
    shape_orig = [ dims[k] for k in dim_order_c ]

    np_img = np_img.reshape(shape_orig).transpose(dim_orig)

    np_img = np_img.squeeze(axis=-1) # remove T

    return np_img


class TokenValidationInterceptor(grpc.ServerInterceptor):
    def __init__(self, token):
        def abort(ignored_request, context):
            context.abort(grpc.StatusCode.UNAUTHENTICATED, "Invalid token signature")

        self._abort_handler = grpc.unary_unary_rpc_method_handler(abort)
        self.token = token

    def intercept_service(self, continuation, handler_call_details):
        expected_metadata = (_AUTH_HEADER_KEY, f"Bearer {self.token}")
        if self.token is None or expected_metadata in handler_call_details.invocation_metadata:
            return continuation(handler_call_details)
        else:
            return self._abort_handler



class LacssServicerBase(proto.LacssServicer):

    def __init__(self):
        self._lock = threading.RLock()


    def RunDetectionStream(self, request_iterator, context):
        with self._lock:
            request = proto.DetectionRequest()

            for next_request in request_iterator:

                if next_request.image_data.HasField("pixels"):
                    request.image_data.pixels.CopyFrom(next_request.image_data.pixels)

                if next_request.image_data.HasField("image_annotation"):
                    request.image_data.image_annotation.CopyFrom(next_request.image_data.image_annotation)
                
                if next_request.HasField("detection_settings"):
                    request.detection_settings.CopyFrom(next_request.detection_settings)
                
                if request.image_data.HasField("pixels"):
                    yield self.RunDetection(request, context)
